import json
from pathlib import Path
from typing import Protocol

from .models import Conversation, ExportData, Message, Tapback


class Serializer(Protocol):
    def serialize(self, data: ExportData) -> str: ...


class JSONSerializer:
    def __init__(self, indent: int | None = 2, include_text: bool = False):
        self.indent = indent
        self.include_text = include_text

    def serialize(self, data: ExportData) -> str:
        payload = {
            "export_date": data.export_date.isoformat(),
            "year": data.year,
            "total_messages": data.total_messages,
            "conversations": {
                key: self._serialize_conversation(conv) for key, conv in data.conversations.items()
            },
        }
        if data.user_name is not None:
            payload["user_name"] = data.user_name
        if data.phrases is not None:
            payload["phrases"] = data.phrases
        if data.sentiment is not None:
            payload["sentiment"] = data.sentiment
        return json.dumps(payload, indent=self.indent, ensure_ascii=False)

    def _serialize_conversation(self, conv: Conversation) -> dict:
        return {
            "chat_identifier": conv.chat_identifier,
            "display_name": conv.display_name,
            "is_group_chat": conv.is_group_chat,
            "participants": conv.participants,
            "messages": [self._serialize_message(msg) for msg in conv.messages],
        }

    def _serialize_message(self, msg: Message) -> dict:
        data = {
            "id": msg.id,
            "guid": msg.guid,
            "timestamp": msg.timestamp_iso,
            "timestamp_unix": msg.timestamp_unix,
            "is_from_me": msg.is_from_me,
            "sender": msg.sender,
            "service": msg.service,
            "has_attachment": msg.has_attachment,
            "text_length": msg.text_length,
            "word_count": msg.word_count,
            "punctuation_count": msg.punctuation_count,
            "has_question": msg.has_question,
            "has_exclamation": msg.has_exclamation,
            "has_link": msg.has_link,
            "emoji_counts": msg.emoji_counts,
        }

        if self.include_text:
            data["text"] = msg.text

        if msg.date_read_after_seconds is not None:
            data["date_read_after_seconds"] = msg.date_read_after_seconds

        if msg.tapbacks:
            data["tapbacks"] = [self._serialize_tapback(tb) for tb in msg.tapbacks]

        return data

    def _serialize_tapback(self, tapback: Tapback) -> dict:
        return {"type": tapback.type, "by": tapback.by}


class JSONLSerializer:
    def __init__(self, include_text: bool = False):
        self.include_text = include_text

    def serialize(self, data: ExportData) -> str:
        lines = []
        for conv_key, conv in data.conversations.items():
            for msg in conv.messages:
                line_data = {
                    "export_date": data.export_date.isoformat(),
                    "year": data.year,
                    "conversation_key": conv_key,
                    "chat_identifier": conv.chat_identifier,
                    "display_name": conv.display_name,
                    "is_group_chat": conv.is_group_chat,
                    "participants": conv.participants,
                    "message": self._serialize_message(msg),
                }
                if data.user_name is not None:
                    line_data["user_name"] = data.user_name
                if data.phrases is not None:
                    line_data["phrases"] = data.phrases
                if data.sentiment is not None:
                    line_data["sentiment"] = data.sentiment
                lines.append(json.dumps(line_data, ensure_ascii=False))
        return "\n".join(lines)

    def _serialize_message(self, msg: Message) -> dict:
        data = {
            "id": msg.id,
            "guid": msg.guid,
            "timestamp": msg.timestamp_iso,
            "timestamp_unix": msg.timestamp_unix,
            "is_from_me": msg.is_from_me,
            "sender": msg.sender,
            "service": msg.service,
            "has_attachment": msg.has_attachment,
            "text_length": msg.text_length,
            "word_count": msg.word_count,
            "punctuation_count": msg.punctuation_count,
            "has_question": msg.has_question,
            "has_exclamation": msg.has_exclamation,
            "has_link": msg.has_link,
            "emoji_counts": msg.emoji_counts,
        }

        if self.include_text:
            data["text"] = msg.text

        if msg.date_read_after_seconds is not None:
            data["date_read_after_seconds"] = msg.date_read_after_seconds

        if msg.tapbacks:
            data["tapbacks"] = [self._serialize_tapback(tb) for tb in msg.tapbacks]

        return data

    def _serialize_tapback(self, tapback: Tapback) -> dict:
        return {"type": tapback.type, "by": tapback.by}


class Exporter:
    def __init__(self, serializer: Serializer | None = None):
        self.serializer = serializer or JSONLSerializer()

    def export_to_string(self, data: ExportData) -> str:
        return self.serializer.serialize(data)

    def export_to_file(self, data: ExportData, output_path: str | Path) -> None:
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        content = self.export_to_string(data)
        output_path.write_text(content, encoding="utf-8")



from .analyzer import (
    NLPStatisticsAnalyzer,
    RawStatisticsAnalyzer,
    StatisticsAnalyzer,
)
from .displays import Display, TerminalDisplay
from .exporter import Exporter, JSONLSerializer, JSONSerializer
from .loader import ExportLoader
from .models import Conversation, ExportData, Message, Tapback
from .permissions import PermissionError, check_database_access, require_database_access
from .service import MessageService

__all__ = [
    "Message",
    "Conversation",
    "ExportData",
    "Tapback",
    "MessageService",
    "Exporter",
    "JSONSerializer",
    "JSONLSerializer",
    "check_database_access",
    "require_database_access",
    "PermissionError",
    "ExportLoader",
    "StatisticsAnalyzer",
    "RawStatisticsAnalyzer",
    "NLPStatisticsAnalyzer",
    "Display",
    "TerminalDisplay",
]



from __future__ import annotations

from collections import Counter
from typing import Any, Dict, List

from .models import Conversation, ExportData, Message
from .sentiment import LexicalSentimentAnalyzer


def _filter_year_messages(conv: Conversation, year: int) -> list[Message]:
    return [
        msg
        for msg in conv.messages
        if not getattr(msg, "is_context_only", False) and msg.timestamp.year == year
    ]


def compute_sentiment_for_export(data: ExportData, interval: str = "month") -> Dict[str, Any]:
    analyzer = LexicalSentimentAnalyzer()

    sent_messages: list[Message] = []
    received_messages: list[Message] = []
    for conv in data.conversations.values():
        for msg in _filter_year_messages(conv, data.year):
            if not (msg.text or "").strip():
                continue
            if msg.is_from_me:
                sent_messages.append(msg)
            else:
                received_messages.append(msg)

    def score_messages(messages: List[Message]) -> Dict[str, Any]:
        distribution = {"positive": 0, "neutral": 0, "negative": 0}
        total_score = 0.0
        total_messages = 0
        period_totals: dict[str, dict[str, Any]] = {}

        for msg in messages:
            text = (msg.text or "").strip()
            if not text:
                continue
            result = analyzer.analyze(text)
            distribution[result.label] += 1
            total_score += result.score
            total_messages += 1

            period_key = _period_key(msg.timestamp, interval)
            period_bucket = period_totals.setdefault(
                period_key,
                {
                    "sum": 0.0,
                    "count": 0,
                    "distribution": {"positive": 0, "neutral": 0, "negative": 0},
                },
            )
            period_bucket["sum"] += result.score
            period_bucket["count"] += 1
            period_bucket["distribution"][result.label] += 1

        return {
            "distribution": distribution,
            "score_sum": total_score,
            "message_count": total_messages,
            "period_totals": period_totals,
        }

    sent_bucket = score_messages(sent_messages)
    received_bucket = score_messages(received_messages)
    overall_bucket = _combine_buckets(sent_bucket, received_bucket)

    if overall_bucket["message_count"] == 0:
        return {}

    sentiment = {
        "overall": _public_view(overall_bucket),
        "sent": _public_view(sent_bucket),
        "received": _public_view(received_bucket),
        "periods": {
            "interval": interval,
            "overall": _format_period_trend(overall_bucket["period_totals"], data.year, interval),
            "sent": _format_period_trend(sent_bucket["period_totals"], data.year, interval),
            "received": _format_period_trend(received_bucket["period_totals"], data.year, interval),
        },
    }
    return sentiment


def _public_view(bucket: dict[str, Any]) -> dict[str, Any]:
    avg_score = (
        round(bucket["score_sum"] / bucket["message_count"], 3) if bucket["message_count"] else 0.0
    )
    distribution = {
        "positive": int(round(bucket["distribution"]["positive"])),
        "neutral": int(round(bucket["distribution"]["neutral"])),
        "negative": int(round(bucket["distribution"]["negative"])),
    }
    return {
        "distribution": distribution,
        "avg_score": avg_score,
        "message_count": int(round(bucket["message_count"])),
    }


def _format_period_trend(
    period_totals: dict[str, dict[str, Any]], year: int, interval: str
) -> list[dict[str, Any]]:
    if interval != "month":
        # Keep the implementation scoped to month granularity for export.
        return []

    trend: list[dict[str, Any]] = []
    for month in range(1, 13):
        period = f"{year}-{month:02d}"
        values = period_totals.get(
            period,
            {
                "sum": 0.0,
                "count": 0,
                "distribution": {"positive": 0, "neutral": 0, "negative": 0},
            },
        )
        count = values["count"]
        if not count:
            trend.append(
                {
                    "period": period,
                    "avg_score": 0.0,
                    "message_count": 0,
                    "distribution": {"positive": 0, "neutral": 0, "negative": 0},
                }
            )
            continue
        avg = round(values["sum"] / count, 3)
        distribution = {
            "positive": int(round(values["distribution"]["positive"])),
            "neutral": int(round(values["distribution"]["neutral"])),
            "negative": int(round(values["distribution"]["negative"])),
        }
        trend.append(
            {
                "period": period,
                "avg_score": avg,
                "message_count": int(round(count)),
                "distribution": distribution,
            }
        )
    return trend


def _combine_buckets(*buckets: dict[str, Any]) -> dict[str, Any]:
    distribution = Counter({"positive": 0, "neutral": 0, "negative": 0})
    total_score = 0.0
    total_messages = 0
    period_totals: dict[str, dict[str, Any]] = {}

    for bucket in buckets:
        for label, count in bucket["distribution"].items():
            distribution[label] += count
        total_score += bucket["score_sum"]
        total_messages += bucket["message_count"]

        for period, values in bucket["period_totals"].items():
            period_bucket = period_totals.setdefault(
                period,
                {
                    "sum": 0.0,
                    "count": 0,
                    "distribution": {"positive": 0, "neutral": 0, "negative": 0},
                },
            )
            period_bucket["sum"] += values["sum"]
            period_bucket["count"] += values["count"]
            for label, count in values["distribution"].items():
                period_bucket["distribution"][label] += count

    return {
        "distribution": dict(distribution),
        "score_sum": total_score,
        "message_count": total_messages,
        "period_totals": period_totals,
    }


def _period_key(timestamp, interval: str) -> str:
    if interval.lower() == "month":
        return timestamp.strftime("%Y-%m")
    raise ValueError(f"Unsupported sentiment interval: {interval}")



from __future__ import annotations

import os
import pwd
from collections import Counter
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from typing import Any

import emoji

APPLE_EPOCH = datetime(2001, 1, 1, tzinfo=timezone.utc)
TIMESTAMP_FACTOR = 1_000_000_000

TAPBACK_MAP = {
    2000: "love",
    2001: "like",
    2002: "dislike",
    2003: "laugh",
    2004: "emphasize",
    2005: "question",
}

EMOJI_JOINERS_AND_MODIFIERS = {
    "\u200d",
    "\ufe0f",
    "\ufe0e",
    "\u20e3",
}

EXCLUDED_EMOJIS = {
    "\ufffc",
    "\u2642",
    "\u2642\ufe0f",
    "\u2640",
    "\u2640\ufe0f",
    "\ufe0f",
}


def get_user_full_name() -> str:
    """
    Get the full name of the current macOS user.
    Falls back to username if full name is not available.
    """
    try:
        username = os.getlogin()
        user_info = pwd.getpwnam(username)
        # GECOS field format: full_name,room,work_phone,home_phone,other
        full_name = user_info.pw_gecos.split(",")[0]
        if full_name and full_name.strip():
            return full_name.strip()
        # Fallback to username
        return username
    except Exception:
        # If anything fails, return a generic fallback
        return "User"


def apple_timestamp_to_datetime(ns: int) -> datetime | None:
    if ns is None or ns == 0:
        return None
    seconds = ns / TIMESTAMP_FACTOR
    return APPLE_EPOCH + timedelta(seconds=seconds)


def datetime_to_apple_timestamp(dt: datetime) -> int:
    seconds = (dt - APPLE_EPOCH).total_seconds()
    return int(seconds * TIMESTAMP_FACTOR)


def calculate_read_duration(date: int, date_read: int) -> float | None:
    if not date_read or date_read <= date:
        return None
    diff_ns = date_read - date
    return diff_ns / TIMESTAMP_FACTOR


def readable_duration(seconds: float) -> str:
    if seconds < 60:
        s = int(seconds)
        return f"{s} second{'s' if s != 1 else ''}"

    parts = []
    hours, remainder = divmod(int(seconds), 3600)
    minutes, secs = divmod(remainder, 60)

    if hours > 0:
        parts.append(f"{hours} hour{'s' if hours != 1 else ''}")
    if minutes > 0:
        parts.append(f"{minutes} minute{'s' if minutes != 1 else ''}")
    if secs > 0 and hours == 0:
        parts.append(f"{secs} second{'s' if secs != 1 else ''}")

    return ", ".join(parts)


def is_tapback(associated_message_type: int | None) -> bool:
    if associated_message_type is None:
        return False
    return 2000 <= associated_message_type <= 2005 or 3000 <= associated_message_type <= 3005


def get_tapback_type(associated_message_type: int) -> str | None:
    return TAPBACK_MAP.get(associated_message_type)


def strip_guid_prefix(guid: str | None) -> str | None:
    if not guid:
        return None
    if "/" in guid:
        return guid.split("/", 1)[1]
    elif ":" in guid:
        return guid.split(":", 1)[1]
    return guid


def extract_text_from_attributed_body(blob: bytes) -> str | None:
    if not blob:
        return None
    try:
        import re

        ns_string_pattern = rb"NSString\x01\x94\x84\x01.(.+?)\x86\x84\x02"
        match = re.search(ns_string_pattern, blob, re.DOTALL)

        if match:
            text_bytes = match.group(1)
            text = text_bytes.decode("utf-8", errors="replace")

            cleaned = "".join(
                c
                for c in text
                if (
                    c.isprintable()
                    or c in "\n\r\t "
                    or ord(c) >= 0x1F300
                    or c in EMOJI_JOINERS_AND_MODIFIERS
                )
                and c != "�"
            )
            if len(cleaned) > 0:
                return cleaned.strip()

        return None
    except Exception:
        return None


def count_emojis(text: str) -> Counter:
    if not text:
        return Counter()

    extracted = emoji.emoji_list(text)
    emojis = []
    for item in extracted:
        emoji_char = item["emoji"]
        if emoji_char in EXCLUDED_EMOJIS:
            continue
        emojis.append(emoji_char)

    return Counter(emojis)


def sanitize_statistics_for_export(statistics: dict[str, Any]) -> dict[str, Any]:
    """
    Return a deep-copied statistics object with privacy-sensitive fields stripped.

    Currently removes per-contact phrase breakdowns so uploads never contain
    potentially identifying content.
    """

    def _strip_private_fields(node: Any) -> None:
        if isinstance(node, dict):
            if "_phrases_by_contact" in node:
                node.pop("_phrases_by_contact", None)
            if "message_distribution" in node and isinstance(node["message_distribution"], list):
                for entry in node["message_distribution"]:
                    if isinstance(entry, dict):
                        entry.pop("contact_name", None)
                        entry.pop("contact_id", None)
            cliff = node.get("cliffhangers")
            if isinstance(cliff, dict):
                cliff.pop("examples", None)
                cliff.pop("examples_them", None)
            if "temporal" in node and isinstance(node["temporal"], dict):
                node["temporal"].pop("weekday_mvp", None)
                node["temporal"].pop("weekend_mvp", None)
            for value in node.values():
                _strip_private_fields(value)
        elif isinstance(node, list):
            for item in node:
                _strip_private_fields(item)

    cleaned = deepcopy(statistics)
    _strip_private_fields(cleaned)
    return cleaned



import logging
import os
import sqlite3
from datetime import datetime, timezone
from typing import Iterator

from .utils import datetime_to_apple_timestamp

logger = logging.getLogger(__name__)


class DatabaseReader:
    def __init__(self, db_path: str | None = None):
        self.db_path = db_path or os.path.expanduser("~/Library/Messages/chat.db")
        self._conn: sqlite3.Connection | None = None

    def connect(self) -> None:
        logger.debug(f"Connecting to database: {self.db_path}")
        self._conn = sqlite3.connect(f"file:{self.db_path}?mode=ro", uri=True)
        self._conn.row_factory = sqlite3.Row
        logger.debug("Database connection established")

    def close(self) -> None:
        if self._conn:
            self._conn.close()
            self._conn = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def get_table_columns(self, table_name: str) -> list[str]:
        assert self._conn is not None, "Database not connected"
        cursor = self._conn.execute(f"PRAGMA table_info({table_name})")
        columns = [row[1] for row in cursor.fetchall()]
        logger.debug(f"Columns in {table_name}: {columns}")
        return columns

    def fetch_messages(self, year: int, batch_size: int = 1000) -> Iterator[dict]:
        logger.debug(f"Fetching messages for year {year}")

        message_columns = self.get_table_columns("message")
        logger.debug(f"Available message columns: {message_columns}")

        start = datetime(year, 1, 1, tzinfo=timezone.utc)
        end = datetime(year, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
        start_ns = datetime_to_apple_timestamp(start)
        end_ns = datetime_to_apple_timestamp(end)

        logger.debug(f"Date range: {start} to {end}")
        logger.debug(f"Timestamp range: {start_ns} to {end_ns}")

        query = """
        SELECT
            m.ROWID as message_id,
            m.guid as message_guid,
            m.text,
            m.attributedBody as attributed_body,
            m.date,
            m.date_read,
            m.is_from_me,
            m.cache_has_attachments,
            m.associated_message_guid,
            m.associated_message_type,
            h.id as sender_id,
            h.service,
            c.ROWID as chat_id,
            c.chat_identifier,
            c.display_name as chat_display_name
        FROM message m
        LEFT JOIN handle h ON m.handle_id = h.ROWID
        LEFT JOIN chat_message_join cmj ON m.ROWID = cmj.message_id
        LEFT JOIN chat c ON cmj.chat_id = c.ROWID
        WHERE m.date >= ? AND m.date <= ?
        ORDER BY m.date ASC
        """

        logger.debug(f"Executing query with parameters: start_ns={start_ns}, end_ns={end_ns}")
        assert self._conn is not None, "Database not connected"
        cursor = self._conn.execute(query, (start_ns, end_ns))
        logger.debug("Query executed successfully")

        while True:
            rows = cursor.fetchmany(batch_size)
            if not rows:
                break
            for row in rows:
                yield dict(row)

    def fetch_chat_participants(self) -> dict[int, list[str]]:
        assert self._conn is not None, "Database not connected"
        query = """
        SELECT chj.chat_id, h.id as participant_id
        FROM chat_handle_join chj
        JOIN handle h ON chj.handle_id = h.ROWID
        ORDER BY chj.chat_id
        """

        cursor = self._conn.execute(query)
        participants = {}

        for row in cursor:
            chat_id = row["chat_id"]
            participant_id = row["participant_id"]
            if chat_id not in participants:
                participants[chat_id] = []
            participants[chat_id].append(participant_id)

        return participants

    def fetch_messages_by_guids(self, guids: list[str]) -> Iterator[dict]:
        if not guids:
            return

        assert self._conn is not None, "Database not connected"
        placeholders = ",".join("?" * len(guids))
        query = f"""
        SELECT
            m.ROWID as message_id,
            m.guid as message_guid,
            m.text,
            m.attributedBody as attributed_body,
            m.date,
            m.date_read,
            m.is_from_me,
            m.cache_has_attachments,
            m.associated_message_guid,
            m.associated_message_type,
            h.id as sender_id,
            h.service,
            c.ROWID as chat_id,
            c.chat_identifier,
            c.display_name as chat_display_name
        FROM message m
        LEFT JOIN handle h ON m.handle_id = h.ROWID
        LEFT JOIN chat_message_join cmj ON m.ROWID = cmj.message_id
        LEFT JOIN chat c ON cmj.chat_id = c.ROWID
        WHERE m.guid IN ({placeholders})
        """

        cursor = self._conn.execute(query, guids)
        for row in cursor:
            yield dict(row)



import logging
import re
from datetime import datetime, timezone

from .db_reader import DatabaseReader
from .models import Conversation, ExportData, Message, Tapback
from .utils import (
    apple_timestamp_to_datetime,
    calculate_read_duration,
    count_emojis,
    extract_text_from_attributed_body,
    get_tapback_type,
    get_user_full_name,
    is_tapback,
    strip_guid_prefix,
)

ME = "Me"
logger = logging.getLogger(__name__)
PUNCTUATION_RE = re.compile(r'[.!?,;:\-\'"()]')
LINK_RE = re.compile(r"https?://", re.IGNORECASE)


class MessageProcessor:
    def __init__(self, reader: DatabaseReader):
        self.reader = reader
        self._guid_to_message = {}

    def process_year(self, year: int) -> ExportData:
        logger.debug(f"Processing messages for year {year}")
        conversations = self._build_conversations(year)
        logger.debug(f"Built {len(conversations)} conversations")

        user_name = get_user_full_name()
        logger.debug(f"Retrieved user name: {user_name}")

        return ExportData(
            export_date=datetime.now(timezone.utc),
            year=year,
            conversations=conversations,
            user_name=user_name,
        )

    def _build_conversations(self, year: int) -> dict[str, Conversation]:
        chat_participants = self.reader.fetch_chat_participants()
        conversations = {}
        message_index = {}
        tapback_queue = []

        for row in self.reader.fetch_messages(year):
            chat_id = row["chat_id"]
            if chat_id is None:
                continue

            chat_key = f"chat_{chat_id}"

            if chat_key not in conversations:
                conversations[chat_key] = self._create_conversation(row, chat_id, chat_participants)

            if is_tapback(row["associated_message_type"]):
                tapback_queue.append(row)
                continue

            message = self._create_message(row)
            if message:
                if message.timestamp.year != year:
                    message.is_context_only = True
                conversations[chat_key].messages.append(message)
                message_index[message.guid] = message

        missing_parent_guids = []
        for tapback_row in tapback_queue:
            parent_guid = strip_guid_prefix(tapback_row["associated_message_guid"])
            if parent_guid and parent_guid not in message_index:
                missing_parent_guids.append(parent_guid)

        if missing_parent_guids:
            unique_missing = list(set(missing_parent_guids))
            logger.debug(
                f"Fetching {len(unique_missing)} unique parent messages for tapbacks from other years"
            )
            logger.debug(f"Sample GUIDs: {unique_missing[:5]}")
            fetched_count = 0
            added_to_index = 0
            tapback_messages = 0
            for row in self.reader.fetch_messages_by_guids(unique_missing):
                fetched_count += 1
                if is_tapback(row["associated_message_type"]):
                    tapback_messages += 1
                    continue
                message = self._create_message(row)
                if message:
                    if message.timestamp.year != year:
                        message.is_context_only = True
                    message_index[message.guid] = message
                    added_to_index += 1

                    chat_id = row["chat_id"]
                    if chat_id:
                        chat_key = f"chat_{chat_id}"
                        if chat_key not in conversations:
                            conversations[chat_key] = self._create_conversation(
                                row, chat_id, chat_participants
                            )
                        conversations[chat_key].messages.append(message)
            logger.debug(
                f"Fetched {fetched_count} rows from DB, skipped {tapback_messages} tapback messages, added {added_to_index} parent messages to index"
            )

        self._apply_tapbacks(tapback_queue, message_index)

        return conversations

    def _create_conversation(
        self, row: dict, chat_id: int, all_participants: dict[int, list[str]]
    ) -> Conversation:
        chat_identifier = row["chat_identifier"] or f"unknown_{chat_id}"
        display_name = row["chat_display_name"]
        participants = all_participants.get(chat_id, [])

        is_group = len(participants) > 1

        return Conversation(
            chat_id=chat_id,
            chat_identifier=chat_identifier,
            display_name=display_name,
            is_group_chat=is_group,
            participants=participants,
        )

    def _create_message(self, row: dict) -> Message | None:
        text = row["text"]
        if not text and row["attributed_body"]:
            text = extract_text_from_attributed_body(row["attributed_body"])

        timestamp = apple_timestamp_to_datetime(row["date"])
        if not timestamp:
            return None

        read_duration = None
        if row["date_read"]:
            read_duration = calculate_read_duration(row["date"], row["date_read"])

        sender = ME if row["is_from_me"] else (row["sender_id"] or "Unknown")
        service = row["service"] or "iMessage"

        text_value = text or ""
        text_length = len(text_value)
        word_count = len(text_value.split()) if text_value.strip() else 0
        punctuation_count = len(PUNCTUATION_RE.findall(text_value)) if text_value else 0
        has_question = "?" in text_value
        has_exclamation = "!" in text_value
        has_link = bool(LINK_RE.search(text_value))
        emojis = count_emojis(text_value)

        return Message(
            id=row["message_id"],
            guid=row["message_guid"],
            timestamp=timestamp,
            is_from_me=bool(row["is_from_me"]),
            sender=sender,
            text=text,
            service=service,
            has_attachment=bool(row["cache_has_attachments"]),
            date_read_after_seconds=read_duration,
            text_length=text_length,
            word_count=word_count,
            punctuation_count=punctuation_count,
            has_question=has_question,
            has_exclamation=has_exclamation,
            has_link=has_link,
            emoji_counts=dict(emojis),
        )

    def _apply_tapbacks(self, tapback_queue: list[dict], message_index: dict[str, Message]) -> None:
        total_tapbacks = len(tapback_queue)
        applied_count = 0
        skipped_no_guid = 0
        skipped_no_type = 0
        skipped_no_parent = 0

        for tapback_row in tapback_queue:
            parent_guid = strip_guid_prefix(tapback_row["associated_message_guid"])
            tapback_type = get_tapback_type(tapback_row["associated_message_type"])

            if not parent_guid:
                skipped_no_guid += 1
                continue

            if not tapback_type:
                skipped_no_type += 1
                continue

            parent_message = message_index.get(parent_guid)
            if not parent_message:
                skipped_no_parent += 1
                continue

            sender = ME if tapback_row["is_from_me"] else (tapback_row["sender_id"] or "Unknown")

            parent_message.tapbacks.append(Tapback(type=tapback_type, by=sender))
            applied_count += 1

        logger.debug(
            f"Tapback application stats: total={total_tapbacks}, applied={applied_count}, "
            f"skipped_no_guid={skipped_no_guid}, skipped_no_type={skipped_no_type}, "
            f"skipped_no_parent={skipped_no_parent}"
        )


class MessageService:
    def __init__(self, db_path: str | None = None):
        self.db_path = db_path

    def export_year(self, year: int) -> ExportData:
        with DatabaseReader(self.db_path) as reader:
            processor = MessageProcessor(reader)
            return processor.process_year(year)



import logging
import re
from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Optional, Sequence

from .ghost import (
    ConversationFilter,
    apply_conversation_filters,
    compute_ghost_stats,
    minimum_responses_filter,
    minimum_total_messages_filter,
    received_to_sent_ratio_filter,
)
from .models import Conversation, ExportData, Message
from .phrases import PhraseExtractionConfig, PhraseExtractor
from .sentiment import LexicalSentimentAnalyzer, SentimentResult
from .utils import count_emojis


class StatisticsAnalyzer(ABC):
    @abstractmethod
    def analyze(self, data: ExportData) -> dict[str, Any]:
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass


logger = logging.getLogger(__name__)

CLIFFHANGER_PATTERNS = [
    "i'll tell you later",
    "ill tell you later",
    "tell you later",
    "wait till you hear",
    "more on that later",
    "remind me to tell you",
    "i'll explain later",
    "ill explain later",
]
CLIFFHANGER_TIMEOUT = timedelta(hours=12)


class RawStatisticsAnalyzer(StatisticsAnalyzer):
    def __init__(
        self,
        sentiment_progress: Optional[Callable[[str, int, int], None]] = None,
        phrase_config: PhraseExtractionConfig | None = None,
        ghost_timeline_days: int = 7,
        ghost_min_consecutive_messages: int = 3,
        ghost_min_conversation_messages: int = 10,
        conversation_filters: Sequence[ConversationFilter] | None = None,
        include_group_chats_in_ghosts: bool = False,
    ) -> None:
        self._sentiment_analyzer = self._build_sentiment_analyzer()
        self._sentiment_interval = "month"
        self._sentiment_progress = sentiment_progress
        self._sentiment_model_info = getattr(self._sentiment_analyzer, "model_info", None)
        self._phrase_extractor = PhraseExtractor(config=phrase_config)

        if ghost_timeline_days <= 0:
            raise ValueError("ghost_timeline_days must be a positive integer")
        self._ghost_timeline = timedelta(days=ghost_timeline_days)
        self._ghost_min_consecutive = ghost_min_consecutive_messages
        self._ghost_min_conversation_messages = ghost_min_conversation_messages
        self._include_group_chats_in_ghosts = include_group_chats_in_ghosts

        if conversation_filters is None:
            conversation_filters = (
                minimum_total_messages_filter(min_total_messages=ghost_min_conversation_messages),
                received_to_sent_ratio_filter(max_ratio=9.0, min_messages_required=2),
                minimum_responses_filter(min_user_responses=2),
            )
        self._conversation_filters: Sequence[ConversationFilter] = conversation_filters

    @property
    def name(self) -> str:
        return "raw"

    @property
    def sentiment_model_info(self) -> dict[str, Any] | None:
        return self._sentiment_model_info

    @staticmethod
    def _to_local_time(dt):
        """Convert UTC datetime to local timezone for temporal analysis."""
        if dt.tzinfo is None:
            return dt
        if dt.tzinfo == timezone.utc:
            return dt.astimezone()
        return dt

    def analyze(self, data: ExportData) -> dict[str, Any]:
        conversations = self._filtered_conversations(data)
        all_messages = self._flatten_messages(data, conversations=conversations)
        all_messages_with_context = self._flatten_messages(
            data, include_context=True, conversations=conversations
        )
        sent_messages = [m for m in all_messages if m.is_from_me]
        received_messages = [m for m in all_messages if not m.is_from_me]
        # Contacts should reflect everyone you messaged or heard from, even if a
        # conversation was filtered out of other analyses. Use the full set.
        contact_conversations = data.conversations

        return {
            "volume": self._analyze_volume(all_messages, sent_messages, received_messages),
            "temporal": self._analyze_temporal_patterns(data, sent_messages, conversations),
            "contacts": self._analyze_contacts(
                data, sent_messages, received_messages, conversations=contact_conversations
            ),
            "content": self._analyze_content(
                data, sent_messages, received_messages, conversations=conversations
            ),
            "conversations": self._analyze_conversations(data, conversations=conversations),
            "response_times": self._analyze_response_times(data, conversations=conversations),
            "tapbacks": self._analyze_tapbacks(
                all_messages_with_context, sent_messages, received_messages
            ),
            "streaks": self._analyze_streaks(data, conversations=conversations),
            "ghosts": self._analyze_ghosts(conversations, data),
            "cliffhangers": self._analyze_cliffhangers(data, conversations),
        }

    def _filtered_conversations(self, data: ExportData) -> dict[str, Conversation]:
        return apply_conversation_filters(
            data.conversations, year=data.year, filters=self._conversation_filters
        )

    def _flatten_messages(
        self,
        data: ExportData,
        include_context: bool = False,
        conversations: dict[str, Conversation] | None = None,
    ) -> list[Message]:
        messages = []
        convs = conversations or data.conversations
        for conv in convs.values():
            for msg in conv.messages:
                if not self._should_include_message(msg, data.year, include_context):
                    continue
                messages.append(msg)
        return sorted(messages, key=lambda m: m.timestamp)

    def _should_include_message(
        self,
        message: Message,
        year: int,
        include_context: bool = False,
    ) -> bool:
        if include_context:
            return True
        if getattr(message, "is_context_only", False):
            return False
        return message.timestamp.year == year

    def _analyze_volume(
        self,
        all_messages: list[Message],
        sent_messages: list[Message],
        received_messages: list[Message],
    ) -> dict[str, Any]:
        sent_by_date = defaultdict(int)
        received_by_date = defaultdict(int)

        for msg in sent_messages:
            local_time = self._to_local_time(msg.timestamp)
            sent_by_date[local_time.date()] += 1

        for msg in received_messages:
            local_time = self._to_local_time(msg.timestamp)
            received_by_date[local_time.date()] += 1

        busiest_day_total = max(
            (
                (date, sent_by_date[date] + received_by_date[date])
                for date in set(sent_by_date.keys()) | set(received_by_date.keys())
            ),
            key=lambda x: x[1],
            default=(None, 0),
        )

        daily_activity = {}
        all_dates = set(sent_by_date.keys()) | set(received_by_date.keys())
        for date in all_dates:
            daily_activity[date.isoformat()] = {
                "sent": sent_by_date.get(date, 0),
                "received": received_by_date.get(date, 0),
                "total": sent_by_date.get(date, 0) + received_by_date.get(date, 0),
            }

        return {
            "total_messages": len(all_messages),
            "total_sent": len(sent_messages),
            "total_received": len(received_messages),
            "busiest_day": {
                "date": busiest_day_total[0].isoformat() if busiest_day_total[0] else None,
                "total": busiest_day_total[1],
                "sent": sent_by_date.get(busiest_day_total[0], 0) if busiest_day_total[0] else 0,
                "received": received_by_date.get(busiest_day_total[0], 0)
                if busiest_day_total[0]
                else 0,
            },
            "most_sent_in_day": max(sent_by_date.values()) if sent_by_date else 0,
            "most_received_in_day": max(received_by_date.values()) if received_by_date else 0,
            "active_days": len(set(sent_by_date.keys()) | set(received_by_date.keys())),
            "days_sent": len(sent_by_date),
            "days_received": len(received_by_date),
            "daily_activity": daily_activity,
        }

    def _analyze_temporal_patterns(
        self,
        data: ExportData,
        sent_messages: list[Message],
        conversations: dict[str, Conversation],
    ) -> dict[str, Any]:
        hour_distribution = Counter(
            self._to_local_time(msg.timestamp).hour for msg in sent_messages
        )
        day_of_week_distribution = Counter(
            self._to_local_time(msg.timestamp).weekday() for msg in sent_messages
        )
        month_distribution = Counter(
            self._to_local_time(msg.timestamp).month for msg in sent_messages
        )

        sorted_hour = dict(sorted(hour_distribution.items()))
        sorted_day = dict(sorted(day_of_week_distribution.items()))

        weekday_counts = sum(count for day, count in sorted_day.items() if day < 5)
        weekend_counts = sum(count for day, count in sorted_day.items() if day >= 5)
        total_sent = sum(sorted_day.values()) or 1

        weekday_contact_counts: dict[str, int] = defaultdict(int)
        weekend_contact_counts: dict[str, int] = defaultdict(int)
        for conv in conversations.values():
            contact_name = conv.display_name or conv.chat_identifier
            messages = self._filter_conversation_messages(conv, data.year)
            for msg in messages:
                if not msg.is_from_me:
                    continue
                local_time = self._to_local_time(msg.timestamp)
                if local_time.weekday() < 5:
                    weekday_contact_counts[contact_name] += 1
                else:
                    weekend_contact_counts[contact_name] += 1

        weekday_mvp = max(
            weekday_contact_counts.items(), key=lambda item: item[1], default=(None, 0)
        )
        weekend_mvp = max(
            weekend_contact_counts.items(), key=lambda item: item[1], default=(None, 0)
        )

        weekday_info = (
            {"contact": weekday_mvp[0], "count": weekday_mvp[1]} if weekday_mvp[0] else None
        )
        weekend_info = (
            {"contact": weekend_mvp[0], "count": weekend_mvp[1]} if weekend_mvp[0] else None
        )

        return {
            "hour_distribution": sorted_hour,
            "day_of_week_distribution": sorted_day,
            "month_distribution": dict(sorted(month_distribution.items())),
            "busiest_hour": hour_distribution.most_common(1)[0] if hour_distribution else (None, 0),
            "weekday_percentage": round(weekday_counts / total_sent * 100, 2),
            "weekend_percentage": round(weekend_counts / total_sent * 100, 2),
            "weekday_mvp": weekday_info,
            "weekend_mvp": weekend_info,
        }

    def _analyze_streaks(
        self,
        data: ExportData,
        conversations: dict[str, Conversation] | None = None,
    ) -> dict[str, Any]:
        max_streak = 0
        max_streak_contact = None
        max_streak_contact_id = None

        convs = conversations or data.conversations
        for conv in convs.values():
            messages = sorted(
                self._filter_conversation_messages(conv, data.year),
                key=lambda msg: msg.timestamp,
            )
            if not messages:
                continue
            messages = sorted(messages, key=lambda m: m.timestamp)

            dates = sorted(set(self._to_local_time(msg.timestamp).date() for msg in messages))

            current_streak = 1
            for i in range(1, len(dates)):
                if dates[i] - dates[i - 1] == timedelta(days=1):
                    current_streak += 1
                    if current_streak > max_streak:
                        max_streak = current_streak
                        max_streak_contact = conv.display_name or conv.chat_identifier
                        max_streak_contact_id = conv.chat_identifier
                else:
                    current_streak = 1

        return {
            "longest_streak_days": max_streak,
            "longest_streak_contact": max_streak_contact,
            "longest_streak_contact_id": max_streak_contact_id,
        }

    def _analyze_cliffhangers(
        self,
        data: ExportData,
        conversations: dict[str, Conversation],
    ) -> dict[str, Any]:
        threshold_hours = int(CLIFFHANGER_TIMEOUT.total_seconds() // 3600)

        def _longest_gap_for_sender(
            ordered_messages: list[Message], *, target_is_from_me: bool, contact_name: str
        ) -> dict[str, Any] | None:
            relevant = [
                msg
                for msg in ordered_messages
                if msg.is_from_me == target_is_from_me and (msg.text or "").strip()
            ]
            if len(relevant) < 2:
                return None

            winner: dict[str, Any] | None = None
            for idx in range(len(relevant) - 1):
                current = relevant[idx]
                next_message = relevant[idx + 1]
                gap = next_message.timestamp - current.timestamp
                if gap <= CLIFFHANGER_TIMEOUT:
                    continue

                snippet = (current.text or "").strip()
                lower_text = snippet.lower()
                matches_pattern = any(pattern in lower_text for pattern in CLIFFHANGER_PATTERNS)
                gap_seconds = gap.total_seconds()
                record = {
                    "contact": contact_name,
                    "timestamp": current.timestamp.isoformat(),
                    "snippet": snippet[:160],
                    "hours_waited": gap_seconds / 3600,
                    "gap_seconds": gap_seconds,
                    "matched_pattern": matches_pattern,
                }

                if winner is None:
                    winner = record
                    continue

                if gap_seconds > winner["gap_seconds"]:
                    winner = record
                    continue

                if (
                    gap_seconds == winner["gap_seconds"]
                    and matches_pattern
                    and not winner["matched_pattern"]
                ):
                    winner = record

            return winner

        example_candidates_you: list[dict[str, Any]] = []
        example_candidates_them: list[dict[str, Any]] = []
        total_you = 0
        total_them = 0

        for conv in conversations.values():
            messages = self._filter_conversation_messages(conv, data.year)
            if not messages:
                continue

            ordered = sorted(messages, key=lambda m: m.timestamp)
            contact_name = conv.display_name or conv.chat_identifier

            longest_you = _longest_gap_for_sender(
                ordered, target_is_from_me=True, contact_name=contact_name
            )
            if longest_you:
                total_you += 1
                example_candidates_you.append(longest_you)

            longest_them = _longest_gap_for_sender(
                ordered, target_is_from_me=False, contact_name=contact_name
            )
            if longest_them:
                total_them += 1
                example_candidates_them.append(longest_them)

        example_candidates_you.sort(key=lambda item: item["gap_seconds"], reverse=True)
        example_candidates_them.sort(key=lambda item: item["gap_seconds"], reverse=True)

        def _build_examples(candidates: list[dict[str, Any]]) -> list[dict[str, Any]]:
            return [
                {
                    "contact": example["contact"],
                    "timestamp": example["timestamp"],
                    "snippet": example["snippet"],
                    "hours_waited": round(example["hours_waited"], 1),
                }
                for example in candidates[:3]
            ]

        longest_wait_you = (
            round(example_candidates_you[0]["hours_waited"], 1) if example_candidates_you else 0.0
        )
        longest_wait_them = (
            round(example_candidates_them[0]["hours_waited"], 1) if example_candidates_them else 0.0
        )

        return {
            "count": total_you,
            "count_them": total_them,
            "threshold_hours": threshold_hours,
            "examples": _build_examples(example_candidates_you),
            "examples_them": _build_examples(example_candidates_them),
            "longest_wait_hours": longest_wait_you,
            "longest_wait_hours_them": longest_wait_them,
        }

    def _analyze_contacts(
        self,
        data: ExportData,
        sent_messages: list[Message],
        received_messages: list[Message],
        conversations: dict[str, Conversation] | None = None,
    ) -> dict[str, Any]:
        sent_by_contact = defaultdict(int)
        received_by_contact = defaultdict(int)
        contact_names = {}

        convs = conversations or data.conversations
        for conv in convs.values():
            contact_id = conv.chat_identifier
            contact_name = conv.display_name or contact_id
            contact_names[contact_id] = contact_name
            messages = self._filter_conversation_messages(conv, data.year)
            if not messages:
                continue
            for msg in messages:
                if msg.is_from_me:
                    sent_by_contact[contact_id] += 1
                else:
                    received_by_contact[contact_id] += 1

        top_sent = sorted(
            ((contact_names[cid], count) for cid, count in sent_by_contact.items()),
            key=lambda x: x[1],
            reverse=True,
        )[:10]

        top_received = sorted(
            ((contact_names[cid], count) for cid, count in received_by_contact.items()),
            key=lambda x: x[1],
            reverse=True,
        )[:10]

        unique_contacts_sent = len([c for c, count in sent_by_contact.items() if count > 0])
        unique_contacts_received = len([c for c, count in received_by_contact.items() if count > 0])

        contacts_by_date_sent = defaultdict(set)
        contacts_by_date_received = defaultdict(set)

        for conv in data.conversations.values():
            contact_id = conv.chat_identifier

            messages = self._filter_conversation_messages(conv, data.year)
            for msg in messages:
                local_time = self._to_local_time(msg.timestamp)
                if msg.is_from_me:
                    contacts_by_date_sent[local_time.date()].add(contact_id)
                else:
                    contacts_by_date_received[local_time.date()].add(contact_id)

        social_butterfly_day = max(
            contacts_by_date_sent.items(), key=lambda x: len(x[1]), default=(None, set())
        )

        fan_club_day = max(
            contacts_by_date_received.items(), key=lambda x: len(x[1]), default=(None, set())
        )

        total_messages_by_contact = {
            cid: sent_by_contact.get(cid, 0) + received_by_contact.get(cid, 0)
            for cid in contact_names.keys()
        }
        total_messages_considered = sum(total_messages_by_contact.values())

        distribution = self._build_chat_concentration(
            total_messages=total_messages_considered,
            totals_by_contact=total_messages_by_contact,
            contact_names=contact_names,
            sent_by_contact=sent_by_contact,
            received_by_contact=received_by_contact,
            top_n=100,
        )

        return {
            "top_sent_to": [{"name": name, "count": count} for name, count in top_sent],
            "top_received_from": [{"name": name, "count": count} for name, count in top_received],
            "unique_contacts_messaged": unique_contacts_sent,
            "unique_contacts_received_from": unique_contacts_received,
            "social_butterfly_day": {
                "date": social_butterfly_day[0].isoformat() if social_butterfly_day[0] else None,
                "unique_contacts": len(social_butterfly_day[1]),
            },
            "fan_club_day": {
                "date": fan_club_day[0].isoformat() if fan_club_day[0] else None,
                "unique_contacts": len(fan_club_day[1]),
            },
            "message_distribution": distribution,
        }

    def _build_chat_concentration(
        self,
        *,
        total_messages: int,
        totals_by_contact: dict[str, int],
        contact_names: dict[str, str],
        sent_by_contact: dict[str, int] | None,
        received_by_contact: dict[str, int] | None,
        top_n: int,
    ) -> list[dict[str, Any]]:
        if not totals_by_contact:
            return []

        sorted_contacts = sorted(
            totals_by_contact.items(),
            key=lambda item: item[1],
            reverse=True,
        )[:top_n]

        denominator = max(total_messages, 1)
        cumulative = 0.0
        distribution = []

        for rank, (contact_id, count) in enumerate(sorted_contacts, start=1):
            share = round(count / denominator, 4)
            cumulative = round(cumulative + share, 4)
            sent_count = (sent_by_contact or {}).get(contact_id, 0)
            received_count = (received_by_contact or {}).get(contact_id, 0)
            sent_ratio = round(sent_count / max(count, 1), 4)
            received_ratio = round(received_count / max(count, 1), 4)
            distribution.append(
                {
                    "rank": rank,
                    "contact_id": contact_id,
                    "contact_name": contact_names.get(contact_id),
                    "share": share,
                    "count": count,
                    "sent_count": sent_count,
                    "received_count": received_count,
                    "sent_ratio": sent_ratio,
                    "received_ratio": received_ratio,
                    "cumulative_share": min(cumulative, 1.0),
                }
            )

        return distribution

    def _analyze_content(
        self,
        data: ExportData,
        sent_messages: list[Message],
        received_messages: list[Message],
        conversations: dict[str, Conversation] | None = None,
    ) -> dict[str, Any]:
        def msg_length(message: Message) -> int:
            if hasattr(message, "text_length"):
                return getattr(message, "text_length") or 0
            return len(message.text or "")

        def msg_word_count(message: Message) -> int:
            if hasattr(message, "word_count"):
                return getattr(message, "word_count") or 0
            text = message.text or ""
            return len(text.split()) if text.strip() else 0

        def msg_punctuation_count(message: Message) -> int:
            if hasattr(message, "punctuation_count"):
                return getattr(message, "punctuation_count") or 0
            return len(re.findall(r'[.!?,;:\-\'"()]', message.text or ""))

        def msg_has_question(message: Message) -> bool:
            if hasattr(message, "has_question"):
                return bool(getattr(message, "has_question"))
            return "?" in (message.text or "")

        def msg_has_exclamation(message: Message) -> bool:
            if hasattr(message, "has_exclamation"):
                return bool(getattr(message, "has_exclamation"))
            return "!" in (message.text or "")

        def msg_has_link(message: Message) -> bool:
            if hasattr(message, "has_link"):
                return bool(getattr(message, "has_link"))
            return bool(re.search(r"https?://", message.text or ""))

        def msg_emoji_counts(message: Message) -> Counter[str]:
            counts = getattr(message, "emoji_counts", None)
            if counts:
                return Counter(counts)
            return count_emojis(message.text or "")

        sent_with_text = [m for m in sent_messages if msg_length(m) > 0]
        received_with_text = [m for m in received_messages if msg_length(m) > 0]

        emoji_counter: Counter[str] = Counter()
        sent_lengths: list[int] = []
        sent_word_counts: list[int] = []
        sent_punctuation_counts: list[int] = []
        received_punctuation_counts: list[int] = []
        question_count = 0
        exclamation_count = 0
        link_count = 0

        for msg in sent_with_text:
            sent_lengths.append(msg_length(msg))
            word_count = msg_word_count(msg)
            sent_word_counts.append(word_count)
            emoji_counter.update(msg_emoji_counts(msg))
            sent_punctuation_counts.append(msg_punctuation_count(msg))
            if msg_has_question(msg):
                question_count += 1
            if msg_has_exclamation(msg):
                exclamation_count += 1
            if msg_has_link(msg):
                link_count += 1

        for msg in received_with_text:
            received_punctuation_counts.append(msg_punctuation_count(msg))

        avg_length_sent = sum(sent_lengths) / len(sent_lengths) if sent_lengths else 0
        avg_length_received = (
            sum(msg_length(m) for m in received_with_text) / len(received_with_text)
            if received_with_text
            else 0
        )

        avg_punctuation_sent = (
            sum(sent_punctuation_counts) / len(sent_punctuation_counts)
            if sent_punctuation_counts
            else 0
        )
        avg_punctuation_received = (
            sum(received_punctuation_counts) / len(received_punctuation_counts)
            if received_punctuation_counts
            else 0
        )

        double_texts = self._count_double_texts(data, conversations)
        if getattr(data, "sentiment", None):
            sentiment_stats = data.sentiment
        else:
            sentiment_stats = self._analyze_sentiment(
                sent_messages,
                received_messages,
                interval=self._sentiment_interval,
            )

        word_count_histogram = self._create_word_count_histogram(sent_word_counts)
        mode_word_count = Counter(sent_word_counts).most_common(1)[0][0] if sent_word_counts else 0
        avg_word_count_sent = (
            sum(sent_word_counts) / len(sent_word_counts) if sent_word_counts else 0
        )

        result = {
            "avg_message_length_sent": round(avg_length_sent, 2),
            "avg_message_length_received": round(avg_length_received, 2),
            "avg_word_count_sent": round(avg_word_count_sent, 2),
            "word_count_histogram": word_count_histogram,
            "mode_word_count": mode_word_count,
            "avg_punctuation_sent": round(avg_punctuation_sent, 2),
            "avg_punctuation_received": round(avg_punctuation_received, 2),
            "most_used_emojis": [
                {"emoji": emoji, "count": count} for emoji, count in emoji_counter.most_common(10)
            ],
            "questions_asked": question_count,
            "questions_percentage": (
                round(question_count / len(sent_with_text) * 100, 2) if sent_with_text else 0
            ),
            "exclamations_sent": exclamation_count,
            "enthusiasm_percentage": (
                round(exclamation_count / len(sent_with_text) * 100, 2) if sent_with_text else 0
            ),
            "links_shared": link_count,
            "attachments_sent": sum(1 for m in sent_messages if m.has_attachment),
            "attachments_received": sum(1 for m in received_messages if m.has_attachment),
            "double_text_count": double_texts["count"],
            "double_text_percentage": double_texts["percentage"],
            "quadruple_text_count": double_texts["quadruple_count"],
        }

        if sentiment_stats:
            result["sentiment"] = sentiment_stats
        phrase_public, phrase_contacts = self._get_phrases(
            data, sent_with_text, conversations=data.conversations
        )
        if phrase_public:
            result["phrases"] = phrase_public
        if phrase_contacts:
            result["_phrases_by_contact"] = phrase_contacts

        return result

    def _count_double_texts(
        self, data: ExportData, conversations: dict[str, Conversation] | None = None
    ) -> dict[str, Any]:
        total_sent = 0
        double_text_count = 0
        quadruple_text_count = 0

        convs = conversations or data.conversations

        for conv in convs.values():
            messages = self._filter_conversation_messages(conv, data.year)
            if not messages:
                continue

            total_sent += sum(1 for msg in messages if msg.is_from_me)

            i = 0
            while i < len(messages):
                current = messages[i]

                if not current.is_from_me:
                    i += 1
                    continue

                run_start_time = current.timestamp
                run_length = 1
                j = i + 1

                while (
                    j < len(messages)
                    and messages[j].is_from_me
                    and (messages[j].timestamp - run_start_time).total_seconds() < 300
                ):
                    run_length += 1
                    j += 1

                if run_length > 1:
                    double_text_count += 1
                if run_length >= 4:
                    quadruple_text_count += 1

                i = j

        percentage = round(double_text_count / total_sent * 100, 2) if total_sent else 0.0

        return {
            "count": double_text_count,
            "percentage": percentage,
            "quadruple_count": quadruple_text_count,
        }

    def _analyze_phrases(
        self,
        data: ExportData,
        sent_messages: list[Message],
        conversations: dict[str, Conversation] | None = None,
    ) -> tuple[dict[str, Any], list[dict[str, Any]]]:
        extractor = getattr(self, "_phrase_extractor", None)
        if extractor is None or not sent_messages:
            return {}, []

        texts = [(msg.text or "").strip() for msg in sent_messages if (msg.text or "").strip()]
        if not texts:
            return {}, []

        per_contact_messages: dict[str, list[str]] = {}
        contact_names: dict[str, str] = {}

        convs = conversations or data.conversations
        for conv in convs.values():
            contact_id = conv.chat_identifier
            contact_names[contact_id] = conv.display_name or contact_id
            per_contact_texts = [
                (msg.text or "").strip()
                for msg in self._filter_conversation_messages(conv, data.year)
                if msg.is_from_me and (msg.text or "").strip()
            ]
            if per_contact_texts:
                per_contact_messages[contact_id] = per_contact_texts

        result = extractor.extract(
            texts,
            per_contact_messages=per_contact_messages or None,
            contact_names=contact_names or None,
        )

        if not result.overall:
            return {}, []

        def serialize_phrase(stat) -> dict[str, Any]:
            value = stat.text
            return {
                "phrase": value,
                "text": value,
                "occurrences": stat.occurrences,
                "share": stat.share,
            }

        overall = [serialize_phrase(stat) for stat in result.overall]
        by_contact = []
        for contact_stats in result.by_contact:
            if not contact_stats.top_phrases:
                continue
            by_contact.append(
                {
                    "contact_id": contact_stats.contact_id,
                    "contact_name": contact_stats.contact_name,
                    "total_messages": contact_stats.total_messages,
                    "top_phrases": [serialize_phrase(stat) for stat in contact_stats.top_phrases],
                }
            )

        config = result.config
        config_info = {
            "ngram_range": list(config.ngram_range),
            "min_occurrences": config.min_occurrences,
            "min_characters": config.min_characters,
            "min_text_messages": config.min_text_messages,
            "per_contact_min_text_messages": config.per_contact_min_text_messages,
            "max_phrases": config.max_phrases,
            "per_contact_limit": config.per_contact_limit,
            "scoring": config.scoring,
        }

        public_payload = {
            "overall": overall,
            "analyzed_messages": result.analyzed_messages,
            "config": config_info,
        }
        return public_payload, by_contact

    def _get_phrases(
        self,
        data: ExportData,
        sent_messages: list[Message],
        conversations: dict[str, Conversation] | None = None,
    ) -> tuple[dict[str, Any], list[dict[str, Any]]]:
        # Prefer phrases precomputed at export time (no raw text needed).
        if getattr(data, "phrases", None) is not None:
            return data.phrases or {}, []

        # Fall back to live extraction only if text is available.
        has_text = any((msg.text or "").strip() for msg in sent_messages)
        if not has_text:
            return {}, []

        return self._analyze_phrases(data, sent_messages, conversations=conversations)

    def _analyze_sentiment(
        self,
        sent_messages: list[Message],
        received_messages: list[Message],
        interval: str = "month",
    ) -> dict[str, Any]:
        sent_bucket = self._score_sentiment_messages(sent_messages, interval, stage="sent")
        received_bucket = self._score_sentiment_messages(
            received_messages, interval, stage="received"
        )
        overall_bucket = self._combine_sentiment_buckets(sent_bucket, received_bucket)

        if overall_bucket["message_count"] == 0:
            return {}

        sentiment = {
            "overall": self._public_sentiment_view(overall_bucket),
            "sent": self._public_sentiment_view(sent_bucket),
            "received": self._public_sentiment_view(received_bucket),
            "periods": {
                "interval": interval,
                "overall": self._format_period_trend(overall_bucket["period_totals"]),
                "sent": self._format_period_trend(sent_bucket["period_totals"]),
                "received": self._format_period_trend(received_bucket["period_totals"]),
            },
        }
        return sentiment

    def _score_sentiment_messages(
        self,
        messages: list[Message],
        interval: str,
        stage: str,
    ) -> dict[str, Any]:
        distribution = {"positive": 0, "neutral": 0, "negative": 0}
        total_score = 0.0
        total_messages = 0.0
        period_totals: dict[str, dict[str, Any]] = {}

        scored_messages = [msg for msg in messages if (msg.text or "").strip()]
        total = len(scored_messages)
        self._report_sentiment_progress(stage, 0, max(total, 1))

        for idx, msg in enumerate(scored_messages, start=1):
            text = (msg.text or "").strip()
            result = self._run_sentiment(text)
            distribution[result.label] += 1
            total_score += result.score
            total_messages += 1

            period_key = self._period_key(msg.timestamp, interval)
            period_bucket = period_totals.setdefault(
                period_key,
                {
                    "sum": 0.0,
                    "count": 0,
                    "distribution": {"positive": 0, "neutral": 0, "negative": 0},
                },
            )
            period_bucket["sum"] += result.score
            period_bucket["count"] += 1
            period_bucket["distribution"][result.label] += 1

            self._report_sentiment_progress(stage, idx, max(total, 1))

        return {
            "distribution": distribution,
            "score_sum": total_score,
            "message_count": total_messages,
            "period_totals": period_totals,
        }

    def _report_sentiment_progress(self, stage: str, completed: int, total: int) -> None:
        if self._sentiment_progress and total:
            try:
                self._sentiment_progress(stage, completed, total)
            except Exception:  # pragma: no cover - best effort progress
                logger.debug("Sentiment progress callback failed", exc_info=True)

    def _clamp_sample_rate(self, value: Any) -> float:
        try:
            rate = float(value)
        except (TypeError, ValueError):
            return 1.0
        rate = max(0.0, min(1.0, rate))
        return rate if rate > 0 else 1.0

    def _run_sentiment(self, text: str) -> SentimentResult:
        return self._sentiment_analyzer.analyze(text)

    def _combine_sentiment_buckets(self, *buckets: dict[str, Any]) -> dict[str, Any]:
        distribution = {"positive": 0, "neutral": 0, "negative": 0}
        total_score = 0.0
        total_messages = 0
        period_totals: dict[str, dict[str, Any]] = {}

        for bucket in buckets:
            for label, count in bucket["distribution"].items():
                distribution[label] += count
            total_score += bucket["score_sum"]
            total_messages += bucket["message_count"]

            for period, values in bucket["period_totals"].items():
                period_bucket = period_totals.setdefault(
                    period,
                    {
                        "sum": 0.0,
                        "count": 0,
                        "distribution": {"positive": 0, "neutral": 0, "negative": 0},
                    },
                )
                period_bucket["sum"] += values["sum"]
                period_bucket["count"] += values["count"]
                for label, count in values["distribution"].items():
                    period_bucket["distribution"][label] += count

        return {
            "distribution": distribution,
            "score_sum": total_score,
            "message_count": total_messages,
            "period_totals": period_totals,
        }

    def _public_sentiment_view(self, bucket: dict[str, Any]) -> dict[str, Any]:
        avg_score = (
            round(bucket["score_sum"] / bucket["message_count"], 3)
            if bucket["message_count"]
            else 0.0
        )

        distribution = {
            "positive": int(round(bucket["distribution"]["positive"])),
            "neutral": int(round(bucket["distribution"]["neutral"])),
            "negative": int(round(bucket["distribution"]["negative"])),
        }

        return {
            "distribution": distribution,
            "avg_score": avg_score,
            "message_count": int(round(bucket["message_count"])),
        }

    def _format_period_trend(
        self, period_totals: dict[str, dict[str, Any]]
    ) -> list[dict[str, Any]]:
        trend = []
        for period in sorted(period_totals.keys()):
            values = period_totals[period]
            count = values["count"]
            if not count:
                continue
            avg = round(values["sum"] / count, 3)
            distribution = {
                "positive": int(round(values["distribution"]["positive"])),
                "neutral": int(round(values["distribution"]["neutral"])),
                "negative": int(round(values["distribution"]["negative"])),
            }
            trend.append(
                {
                    "period": period,
                    "avg_score": avg,
                    "message_count": int(round(count)),
                    "distribution": distribution,
                }
            )
        return trend

    def _period_key(self, timestamp: datetime, interval: str) -> str:
        interval = interval.lower()
        if interval == "month":
            return timestamp.strftime("%Y-%m")
        if interval == "week":
            iso = timestamp.isocalendar()
            return f"{iso.year}-W{iso.week:02d}"
        if interval == "day":
            return timestamp.strftime("%Y-%m-%d")
        raise ValueError(f"Unsupported sentiment interval: {interval}")

    def _build_sentiment_analyzer(self):
        return LexicalSentimentAnalyzer()

    def _filter_conversation_messages(
        self,
        conversation: Conversation,
        year: int,
    ) -> list[Message]:
        return [
            msg
            for msg in conversation.messages
            if self._should_include_message(msg, year, include_context=False)
        ]

    def _create_word_count_histogram(self, word_counts: list[int]) -> dict[str, int]:
        if not word_counts:
            return {}

        histogram = Counter()
        for count in word_counts:
            histogram[count] += 1

        return dict(histogram)

    def _analyze_conversations(
        self,
        data: ExportData,
        conversations: dict[str, Conversation] | None = None,
    ) -> dict[str, Any]:
        convs = conversations or data.conversations
        group_chats = [c for c in convs.values() if c.is_group_chat]
        one_on_one = [c for c in convs.values() if not c.is_group_chat]

        group_message_count = sum(c.message_count for c in group_chats)
        one_on_one_message_count = sum(c.message_count for c in one_on_one)
        total = group_message_count + one_on_one_message_count

        most_active = max(convs.values(), key=lambda c: c.message_count, default=None)

        most_active_group = (
            max(group_chats, key=lambda c: c.message_count, default=None) if group_chats else None
        )

        return {
            "total_conversations": len(convs),
            "group_chats": len(group_chats),
            "one_on_one_chats": len(one_on_one),
            "group_vs_1on1_ratio": {
                "group_percentage": round(group_message_count / total * 100, 2) if total > 0 else 0,
                "one_on_one_percentage": round(one_on_one_message_count / total * 100, 2)
                if total > 0
                else 0,
            },
            "most_active_thread": {
                "name": most_active.display_name or most_active.chat_identifier
                if most_active
                else None,
                "message_count": most_active.message_count if most_active else 0,
                "is_group": most_active.is_group_chat if most_active else False,
            },
            "most_active_group_chat": {
                "name": most_active_group.display_name or most_active_group.chat_identifier
                if most_active_group
                else None,
                "message_count": most_active_group.message_count if most_active_group else 0,
            }
            if most_active_group
            else None,
        }

    def _analyze_ghosts(
        self,
        conversations: dict[str, Conversation],
        data: ExportData,
    ) -> dict[str, Any]:
        stats = compute_ghost_stats(
            conversations.values(),
            year=data.year,
            timeline=self._ghost_timeline,
            min_consecutive_messages=self._ghost_min_consecutive,
            min_conversation_messages=self._ghost_min_conversation_messages,
            reference_time=data.export_date,
            include_group_chats=self._include_group_chats_in_ghosts,
        )
        ratio = None
        if stats.ghosted_you_count:
            ratio = round(stats.you_ghosted_count / stats.ghosted_you_count, 2)

        timeline_days = max(int(self._ghost_timeline.total_seconds() // 86400), 1)

        return {
            "timeline_days": timeline_days,
            "min_consecutive_messages": self._ghost_min_consecutive,
            "min_conversation_messages": self._ghost_min_conversation_messages,
            "people_you_left_hanging": stats.you_ghosted_count,
            "people_who_left_you_hanging": stats.ghosted_you_count,
            "ghost_ratio": ratio,
        }

    def _analyze_response_times(
        self,
        data: ExportData,
        conversations: dict[str, Conversation] | None = None,
    ) -> dict[str, Any]:
        response_times_you = []
        response_times_them = []

        convs = conversations or data.conversations
        for conv in convs.values():
            # Skip group chats - response dynamics are different
            if conv.is_group_chat:
                continue
            messages = self._filter_conversation_messages(conv, data.year)
            if len(messages) < 2:
                continue
            messages = sorted(messages, key=lambda m: m.timestamp)

            for i in range(len(messages) - 1):
                current = messages[i]
                next_msg = messages[i + 1]

                time_diff_seconds = (next_msg.timestamp - current.timestamp).total_seconds()

                if current.is_from_me and not next_msg.is_from_me:
                    response_times_them.append(time_diff_seconds)
                elif not current.is_from_me and next_msg.is_from_me:
                    response_times_you.append(time_diff_seconds)

        def format_duration(seconds: float) -> str:
            if seconds < 60:
                return f"{int(seconds)}s"
            elif seconds < 3600:
                return f"{int(seconds // 60)}m {int(seconds % 60)}s"
            elif seconds < 86400:
                hours = int(seconds // 3600)
                minutes = int((seconds % 3600) // 60)
                return f"{hours}h {minutes}m"
            else:
                days = int(seconds // 86400)
                hours = int((seconds % 86400) // 3600)
                return f"{days}d {hours}h"

        def calculate_median(times: list[float]) -> float:
            if not times:
                return 0
            sorted_times = sorted(times)
            n = len(sorted_times)
            if n % 2 == 0:
                return (sorted_times[n // 2 - 1] + sorted_times[n // 2]) / 2
            return sorted_times[n // 2]

        median_response_you = calculate_median(response_times_you)
        median_response_them = calculate_median(response_times_them)

        return {
            "median_response_time_you_seconds": round(median_response_you, 2),
            "median_response_time_you_formatted": format_duration(median_response_you),
            "median_response_time_them_seconds": round(median_response_them, 2),
            "median_response_time_them_formatted": format_duration(median_response_them),
            "total_responses_you": len(response_times_you),
            "total_responses_them": len(response_times_them),
        }

    def _analyze_tapbacks(
        self,
        all_messages: list[Message],
        sent_messages: list[Message],
        received_messages: list[Message],
    ) -> dict[str, Any]:
        tapbacks_given = []
        tapbacks_received = []

        for msg in all_messages:
            for tapback in msg.tapbacks:
                if tapback.by == "Me":
                    tapbacks_given.append(tapback.type)
                else:
                    tapbacks_received.append(tapback.type)

        tapback_counter_given = Counter(tapbacks_given)
        tapback_counter_received = Counter(tapbacks_received)

        all_tapback_types = ["love", "like", "dislike", "laugh", "emphasize", "question"]
        tapback_distribution_given = {t: tapback_counter_given.get(t, 0) for t in all_tapback_types}
        tapback_distribution_received = {
            t: tapback_counter_received.get(t, 0) for t in all_tapback_types
        }

        return {
            "total_tapbacks_given": len(tapbacks_given),
            "total_tapbacks_received": len(tapbacks_received),
            "favorite_tapback": tapback_counter_given.most_common(1)[0]
            if tapback_counter_given
            else (None, 0),
            "most_received_tapback": tapback_counter_received.most_common(1)[0]
            if tapback_counter_received
            else (None, 0),
            "tapback_distribution_given": tapback_distribution_given,
            "tapback_distribution_received": tapback_distribution_received,
        }


class NLPStatisticsAnalyzer(StatisticsAnalyzer):
    @property
    def name(self) -> str:
        return "nlp"

    def analyze(self, data: ExportData) -> dict[str, Any]:
        return {
            "status": "not_implemented",
            "message": "NLP analysis requires additional dependencies (spaCy, NLTK)",
            "planned_features": [
                "sentiment_analysis",
                "topic_clustering",
                "word_frequency_analysis",
                "linguistic_patterns",
                "named_entity_extraction",
            ],
        }



from typing import Any, Tuple

from .models import Conversation, ExportData, Message
from .phrases import PhraseExtractionConfig, PhraseExtractor


def _filter_year_messages(conversation: Conversation, year: int) -> list[Message]:
    return [
        msg
        for msg in conversation.messages
        if not getattr(msg, "is_context_only", False) and msg.timestamp.year == year
    ]


def compute_phrases_for_export(
    data: ExportData, phrase_config: PhraseExtractionConfig | None = None
) -> Tuple[dict[str, Any], list[dict[str, Any]]]:
    extractor = PhraseExtractor(config=phrase_config)

    texts = []
    for conv in data.conversations.values():
        conv_texts = []
        for msg in _filter_year_messages(conv, data.year):
            text = (msg.text or "").strip()
            if not text or not msg.is_from_me:
                continue
            texts.append(text)
            conv_texts.append(text)

    if not texts:
        return {}, []

    result = extractor.extract(
        texts,
        per_contact_messages=None,
        contact_names=None,
    )

    if not result.overall:
        return {}, []

    def serialize_phrase(stat) -> dict[str, Any]:
        value = stat.text
        return {
            "phrase": value,
            "text": value,
            "occurrences": stat.occurrences,
            "share": stat.share,
        }

    overall = [serialize_phrase(stat) for stat in result.overall][:10]

    config = result.config
    config_info = {
        "ngram_range": list(config.ngram_range),
        "min_occurrences": config.min_occurrences,
        "min_characters": config.min_characters,
        "min_text_messages": config.min_text_messages,
        "per_contact_min_text_messages": config.per_contact_min_text_messages,
        "max_phrases": config.max_phrases,
        "per_contact_limit": config.per_contact_limit,
        "scoring": config.scoring,
    }

    public_payload = {
        "overall": overall,
        "analyzed_messages": result.analyzed_messages,
        "config": config_info,
    }

    return public_payload, []



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Iterable

from ..models import Conversation, Message


@dataclass
class GhostStats:
    timeline: timedelta
    reference_time: datetime
    min_consecutive_messages: int
    min_conversation_messages: int
    you_ghosted_count: int = 0
    ghosted_you_count: int = 0


def compute_ghost_stats(
    conversations: Iterable[Conversation],
    *,
    year: int,
    timeline: timedelta,
    min_consecutive_messages: int,
    min_conversation_messages: int,
    reference_time: datetime | None = None,
    include_group_chats: bool = False,
) -> GhostStats:
    """
    Classify which contacts were ghosted in each direction.

    Args:
        conversations: Conversations to evaluate (already filtered upstream).
        year: The export year to respect when reading messages.
        timeline: Maximum allowed silence between replies.
        min_consecutive_messages: Minimum run length before silence counts.
        min_conversation_messages: Minimum sent+received messages for analysis.
        reference_time: Used when no further responses exist.
        include_group_chats: Whether to include group conversations.
    """

    if timeline <= timedelta(0):
        raise ValueError("timeline must be a positive duration")

    if min_consecutive_messages <= 0:
        raise ValueError("min_consecutive_messages must be positive")

    reference_time = _normalize_reference_time(reference_time)
    stats = GhostStats(
        timeline=timeline,
        reference_time=reference_time,
        min_consecutive_messages=min_consecutive_messages,
        min_conversation_messages=min_conversation_messages,
    )

    for conversation in conversations:
        if conversation.is_group_chat and not include_group_chats:
            continue

        messages = _conversation_messages(conversation, year)
        if len(messages) < min_conversation_messages:
            continue

        you_ghosted, they_ghosted = _classify_conversation(
            messages,
            timeline,
            reference_time,
            min_consecutive_messages=min_consecutive_messages,
        )

        if you_ghosted:
            stats.you_ghosted_count += 1
        if they_ghosted:
            stats.ghosted_you_count += 1

    return stats


def _normalize_reference_time(ts: datetime | None) -> datetime:
    if ts is None:
        return datetime.now(timezone.utc)
    if ts.tzinfo is None:
        return ts.replace(tzinfo=timezone.utc)
    return ts


def _conversation_messages(conversation: Conversation, year: int) -> list[Message]:
    messages = [
        msg
        for msg in conversation.messages
        if not getattr(msg, "is_context_only", False) and msg.timestamp.year == year
    ]
    messages.sort(key=lambda m: m.timestamp)
    return messages


def _classify_conversation(
    messages: list[Message],
    timeline: timedelta,
    reference_time: datetime,
    *,
    min_consecutive_messages: int,
) -> tuple[bool, bool]:
    """
    Determine whether each direction satisfies the ghost criteria.
    """

    runs = _generate_runs(messages)
    you_ghosted = False
    ghosted_you = False

    for run in runs:
        if run.length < min_consecutive_messages:
            continue

        response_time = _next_response_time(messages, run.end_index, reference_time)
        if response_time is None:
            continue

        silence = response_time - run.end_timestamp
        if silence < timeline:
            continue

        if run.actor == "me":
            ghosted_you = True
        else:
            you_ghosted = True

        if you_ghosted and ghosted_you:
            break

    return you_ghosted, ghosted_you


@dataclass
class _Run:
    actor: str  # "me" or "them"
    length: int
    end_index: int
    end_timestamp: datetime


def _generate_runs(messages: list[Message]) -> list[_Run]:
    runs: list[_Run] = []
    current_actor: str | None = None
    run_length = 0
    for idx, message in enumerate(messages):
        actor = "me" if message.is_from_me else "them"
        if actor != current_actor:
            if current_actor is not None:
                runs.append(
                    _Run(
                        actor=current_actor,
                        length=run_length,
                        end_index=idx - 1,
                        end_timestamp=messages[idx - 1].timestamp,
                    )
                )
            current_actor = actor
            run_length = 1
        else:
            run_length += 1

    if current_actor is not None and run_length > 0:
        runs.append(
            _Run(
                actor=current_actor,
                length=run_length,
                end_index=len(messages) - 1,
                end_timestamp=messages[-1].timestamp,
            )
        )

    return runs


def _next_response_time(
    messages: list[Message],
    run_end_index: int,
    reference_time: datetime,
) -> datetime | None:
    if run_end_index == len(messages) - 1:
        return reference_time
    return messages[run_end_index + 1].timestamp



from datetime import datetime
from typing import Any

from rich.columns import Columns
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from .base import Display


class TerminalDisplay(Display):
    def __init__(self):
        self.console = Console()

    def render(self, statistics: dict[str, Any], brief: bool = False) -> None:
        if brief:
            self.render_brief(statistics)
        else:
            self.console.print()
            self._render_header(statistics)

            if "raw" in statistics:
                self._render_raw_statistics(statistics["raw"])

            if "nlp" in statistics:
                self._render_stub_section("NLP Analysis", statistics["nlp"])

            if "llm" in statistics:
                self._render_stub_section("LLM Analysis", statistics["llm"])

            self.console.print()

    def render_brief(self, statistics: dict[str, Any]) -> None:
        self.console.print()

        if "raw" not in statistics:
            return

        stats = statistics["raw"]
        volume = stats.get("volume", {})
        contacts = stats.get("contacts", {})
        content = stats.get("content", {})

        title = Text("Your iMessage Year in Review (Summary)", style="bold magenta")
        self.console.print(Panel(title, border_style="magenta"))

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("", style="dim", width=25)
        table.add_column("", style="bold cyan")

        table.add_row("📊 Total Messages", f"{volume.get('total_messages', 0):,}")
        table.add_row("💬 Messages Sent", f"{volume.get('total_sent', 0):,}")
        table.add_row("📥 Messages Received", f"{volume.get('total_received', 0):,}")

        top_contacts = contacts.get("top_sent_to", [])
        if top_contacts:
            top_contact = top_contacts[0]
            table.add_row(
                "👤 Top Contact", f"{top_contact['name']} ({top_contact['count']:,} msgs)"
            )

        emojis = content.get("most_used_emojis", [])
        if emojis:
            top_emoji = emojis[0]
            table.add_row(
                "😊 Favorite Emoji", f"{top_emoji['emoji']} ({top_emoji['count']:,} times)"
            )

        self.console.print(table)
        self.console.print()

    def _render_header(self, statistics: dict[str, Any]) -> None:
        title = Text("iMessage Year in Review 2025", style="bold magenta")
        self.console.print(Panel(title, border_style="magenta"))

    def _render_raw_statistics(self, stats: dict[str, Any]) -> None:
        self._render_volume_section(stats.get("volume", {}))
        self._render_temporal_section(stats.get("temporal", {}))
        self._render_streaks_section(stats.get("streaks", {}))
        self._render_contacts_section(stats.get("contacts", {}))
        self._render_content_section(stats.get("content", {}))
        self._render_conversations_section(stats.get("conversations", {}))
        self._render_response_times_section(stats.get("response_times", {}))
        self._render_tapbacks_section(stats.get("tapbacks", {}))
        self._render_ghosts_section(stats.get("ghosts", {}))
        self._render_cliffhangers_section(stats.get("cliffhangers"))

    def _render_volume_section(self, volume: dict[str, Any]) -> None:
        self.console.print("\n[bold cyan]📊 Volume & Activity[/]")

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("Metric", style="dim")
        table.add_column("Value", style="green")

        table.add_row("Total Messages", f"{volume.get('total_messages', 0):,}")
        table.add_row("Messages Sent", f"{volume.get('total_sent', 0):,}")
        table.add_row("Messages Received", f"{volume.get('total_received', 0):,}")
        table.add_row("Active Days", f"{volume.get('active_days', 0):,}")

        busiest = volume.get("busiest_day", {})
        if busiest.get("date"):
            table.add_row("Busiest Day", f"{busiest['date']} ({busiest['total']:,} messages)")

        table.add_row("Most Sent in One Day", f"{volume.get('most_sent_in_day', 0):,}")
        table.add_row("Most Received in One Day", f"{volume.get('most_received_in_day', 0):,}")

        self.console.print(table)

    def _render_temporal_section(self, temporal: dict[str, Any]) -> None:
        self.console.print("\n[bold cyan]⏰ Temporal Patterns[/]")

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("Metric", style="dim")
        table.add_column("Value", style="yellow")

        busiest_hour = temporal.get("busiest_hour", (None, 0))
        if busiest_hour[0] is not None:
            hour_12 = busiest_hour[0] % 12 or 12
            am_pm = "AM" if busiest_hour[0] < 12 else "PM"
            table.add_row("Busiest Hour", f"{hour_12}:00 {am_pm} ({busiest_hour[1]:,} messages)")

        self.console.print(table)

        self._render_weekday_weekend_breakdown(temporal)

    def _render_weekday_weekend_breakdown(self, temporal: dict[str, Any]) -> None:
        weekday_pct = temporal.get("weekday_percentage")
        weekend_pct = temporal.get("weekend_percentage")
        if weekday_pct is None or weekend_pct is None:
            return

        segments = 12
        filled_segments = int(round(segments * (weekday_pct / 100)))
        filled_segments = max(0, min(segments, filled_segments))
        characters = ["●" if i < filled_segments else "○" for i in range(segments)]
        characters = [
            f"[magenta]{char}[/]" if char == "●" else f"[dim]{char}[/]" for char in characters
        ]
        layout = [
            f"    {characters[0]} {characters[1]} {characters[2]} {characters[3]}    ",
            f"  {characters[11]}           {characters[4]}  ",
            f"  {characters[10]}           {characters[5]}  ",
            f"    {characters[9]} {characters[8]} {characters[7]} {characters[6]}    ",
        ]

        self.console.print("\n[bold]Weekday vs Weekend[/]")
        for line in layout:
            self.console.print(line)

        weekday_info = temporal.get("weekday_mvp") or {}
        weekend_info = temporal.get("weekend_mvp") or {}

        weekday_contact = weekday_info.get("contact") or "—"
        weekend_contact = weekend_info.get("contact") or "—"
        weekday_count = weekday_info.get("count")
        weekend_count = weekend_info.get("count")

        weekday_suffix = f" ({(weekday_count or 0):,} msgs)" if weekday_count is not None else ""
        weekend_suffix = f" ({(weekend_count or 0):,} msgs)" if weekend_count is not None else ""

        self.console.print(
            f"[bold magenta]Weekday Warrior:[/] {weekday_contact} — {weekday_pct:.1f}%{weekday_suffix}"
        )
        self.console.print(
            f"[bold cyan]Weekend MVP:[/] {weekend_contact} — {weekend_pct:.1f}%{weekend_suffix}"
        )

    def _render_streaks_section(self, streaks: dict[str, Any]) -> None:
        if streaks.get("longest_streak_days", 0) == 0:
            return

        self.console.print("\n[bold cyan]🔥 Streaks[/]")

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("Metric", style="dim")
        table.add_column("Value", style="yellow")

        streak_days = streaks.get("longest_streak_days", 0)
        streak_contact = streaks.get("longest_streak_contact", "Unknown")
        table.add_row("Longest Streak", f"{streak_days} days with {streak_contact}")

        self.console.print(table)

    def _render_contacts_section(self, contacts: dict[str, Any]) -> None:
        self.console.print("\n[bold cyan]👥 Top Contacts[/]")

        top_sent = contacts.get("top_sent_to", [])[:5]
        top_received = contacts.get("top_received_from", [])[:5]

        if top_sent:
            self.console.print("\n[bold]Most Messaged:[/]")
            table = Table(show_header=False, box=None, padding=(0, 2))
            table.add_column("Rank", style="dim", width=4)
            table.add_column("Contact", style="cyan")
            table.add_column("Count", style="green", justify="right")

            for i, contact in enumerate(top_sent, 1):
                table.add_row(f"{i}.", contact["name"], f"{contact['count']:,}")

            self.console.print(table)

        if top_received:
            self.console.print("\n[bold]Most Received From:[/]")
            table = Table(show_header=False, box=None, padding=(0, 2))
            table.add_column("Rank", style="dim", width=4)
            table.add_column("Contact", style="cyan")
            table.add_column("Count", style="green", justify="right")

            for i, contact in enumerate(top_received, 1):
                table.add_row(f"{i}.", contact["name"], f"{contact['count']:,}")

            self.console.print(table)

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("Metric", style="dim")
        table.add_column("Value", style="magenta")

        table.add_row("Unique Contacts Messaged", f"{contacts.get('unique_contacts_messaged', 0)}")
        table.add_row(
            "Unique Contacts Received From", f"{contacts.get('unique_contacts_received_from', 0)}"
        )

        butterfly = contacts.get("social_butterfly_day", {})
        if butterfly.get("date"):
            table.add_row(
                "Social Butterfly Day",
                f"{butterfly['date']} ({butterfly['unique_contacts']} people)",
            )

        fan_club = contacts.get("fan_club_day", {})
        if fan_club.get("date"):
            table.add_row(
                "Fan Club Day", f"{fan_club['date']} ({fan_club['unique_contacts']} people)"
            )

        self.console.print()
        self.console.print(table)

        distribution = contacts.get("message_distribution") or []
        if distribution:
            self.console.print("\n[bold]Chat Concentration[/]")
            max_rows = min(10, len(distribution))
            for entry in distribution[:max_rows]:
                share = entry.get("share", 0.0)
                bar = self._share_bar(share)
                self.console.print(f"{entry.get('rank', 0):>2}. {bar}")

    def _render_content_section(self, content: dict[str, Any]) -> None:
        self.console.print("\n[bold cyan]💬 Message Content[/]")

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("Metric", style="dim")
        table.add_column("Value", style="blue")

        table.add_row(
            "Avg Message Length (Sent)", f"{content.get('avg_message_length_sent', 0)} chars"
        )
        table.add_row(
            "Avg Message Length (Received)",
            f"{content.get('avg_message_length_received', 0)} chars",
        )
        table.add_row("Questions Asked", f"{content.get('questions_percentage', 0)}%")
        table.add_row("Enthusiasm Level", f"{content.get('enthusiasm_percentage', 0)}%")
        table.add_row("Attachments Sent", f"{content.get('attachments_sent', 0):,}")
        table.add_row("Attachments Received", f"{content.get('attachments_received', 0):,}")
        table.add_row(
            "Double Text Count",
            f"{content.get('double_text_count', 0):,} ({content.get('double_text_percentage', 0)}%)",
        )
        table.add_row(
            "Quadruple Text Count (Crash Outs)",
            f"{content.get('quadruple_text_count', 0):,}",
        )

        self.console.print(table)

        emojis = content.get("most_used_emojis", [])[:5]
        if emojis:
            self.console.print("\n[bold]Most Used Emojis:[/]")
            emoji_texts = [Text(f"{e['emoji']} {e['count']:,}", style="yellow") for e in emojis]
            self.console.print(Columns(emoji_texts, padding=(0, 3)))

        phrases = content.get("phrases")
        by_contact = content.get("_phrases_by_contact") or []
        if phrases:
            self._render_phrase_section(phrases, by_contact)

        sentiment = content.get("sentiment")
        if sentiment:
            self._render_sentiment_overview(sentiment)
            self._render_sentiment_periods(sentiment.get("periods"))

    def _render_phrase_section(
        self, phrases: dict[str, Any], by_contact: list[dict[str, Any]]
    ) -> None:
        overall = phrases.get("overall") or []
        if overall:
            self.console.print("\n[bold]Most Used Phrases:[/]")
            table = Table(show_header=True, header_style="bold green")
            table.add_column("Rank", style="dim", width=4)
            table.add_column("Phrase", style="white")
            table.add_column("Count", style="green", justify="right")

            for idx, phrase in enumerate(overall, start=1):
                label = phrase.get("phrase") or phrase.get("text") or "—"
                table.add_row(f"{idx}.", label, f"{phrase.get('occurrences', 0):,}")

        self.console.print(table)

        filtered = [entry for entry in by_contact if entry.get("top_phrases")]
        if filtered:
            self.console.print("\n[bold]Signature Lines by Contact:[/]")
            table = Table(show_header=True, header_style="bold blue")
            table.add_column("Contact", style="cyan")
            table.add_column("Phrase", style="white")
            table.add_column("Count", style="green", justify="right")

            for entry in filtered[:3]:
                top = entry["top_phrases"][0]
                label = top.get("phrase") or top.get("text") or "—"
                contact_name = entry.get("contact_name") or entry.get("contact_id")
                table.add_row(contact_name, label, f"{top.get('occurrences', 0):,}")

            self.console.print(table)

    def _render_sentiment_overview(self, sentiment: dict[str, Any]) -> None:
        self.console.print("\n[bold cyan]🧠 Sentiment Snapshot[/]")
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Perspective", style="dim")
        table.add_column("Positive", justify="right")
        table.add_column("Neutral", justify="right")
        table.add_column("Negative", justify="right")
        table.add_column("Avg Score", justify="right")

        perspectives = [
            ("Overall", sentiment.get("overall")),
            ("You", sentiment.get("sent")),
            ("Them", sentiment.get("received")),
        ]

        has_data = False
        for label, data in perspectives:
            if not data or data.get("message_count", 0) == 0:
                continue
            has_data = True
            percentages = self._sentiment_percentages(data.get("distribution", {}))
            avg_score = data.get("avg_score", 0.0)
            table.add_row(
                label,
                percentages["positive"],
                percentages["neutral"],
                percentages["negative"],
                f"{avg_score:+.2f}",
            )

        if has_data:
            self.console.print(table)
        else:
            self.console.print("[dim]No sentiment-ready messages found.[/]")

    def _render_sentiment_periods(self, periods: dict[str, Any] | None) -> None:
        if not periods:
            return

        interval = periods.get("interval")
        if interval != "month":
            return

        sent = periods.get("sent", [])
        received = periods.get("received", [])
        rows = self._merge_period_rows(sent, received)
        if not rows:
            return

        self.console.print("\n[bold cyan]📅 Monthly Sentiment (You vs Them)[/]")
        table = Table(show_header=True, header_style="bold cyan")
        table.add_column("Month", style="dim")
        table.add_column("You (avg)", justify="right")
        table.add_column("Them (avg)", justify="right")

        for row in rows:
            table.add_row(
                self._format_period_label(row["period"], interval),
                self._format_sentiment_value(row["sent"]),
                self._format_sentiment_value(row["received"]),
            )

        self.console.print(table)

        overall = periods.get("overall") or []
        if overall:
            self._render_sentiment_bar_chart(overall, interval)

    def _share_bar(self, share: float) -> str:
        total_width = 30
        filled = max(1, min(total_width, int(round(share * total_width))))
        empty = total_width - filled
        return f"[magenta]{'█' * filled}[/][dim]{'·' * empty}[/]"

    def _render_sentiment_bar_chart(self, rows: list[dict[str, Any]], interval: str) -> None:
        if not rows:
            return

        self.console.print("\n[bold cyan]🌈 Mood Flow[/]")
        axis_label = "Month" if interval == "month" else interval.title()
        self.console.print(f"[dim]{axis_label} vs sentiment intensity[/]")

        for row in rows:
            score = row.get("avg_score", 0.0)
            label = self._format_period_label(row["period"], interval)
            magnitude = min(1.0, abs(score))
            width = max(1, int(magnitude * 20))
            color = "magenta" if score >= 0 else "cyan"
            bar = "█" * width
            self.console.print(f"{label:>7} [bold {color}]{bar}[/]")

    def _render_conversations_section(self, conversations: dict[str, Any]) -> None:
        self.console.print("\n[bold cyan]💭 Conversations[/]")

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("Metric", style="dim")
        table.add_column("Value", style="cyan")

        table.add_row("Total Conversations", f"{conversations.get('total_conversations', 0):,}")
        table.add_row("Group Chats", f"{conversations.get('group_chats', 0):,}")
        table.add_row("1-on-1 Chats", f"{conversations.get('one_on_one_chats', 0):,}")

        ratio = conversations.get("group_vs_1on1_ratio", {})
        table.add_row(
            "Group vs 1:1",
            f"{ratio.get('group_percentage', 0)}% / {ratio.get('one_on_one_percentage', 0)}%",
        )

        most_active = conversations.get("most_active_thread", {})
        if most_active.get("name"):
            chat_type = "Group" if most_active.get("is_group") else "1:1"
            table.add_row(
                "Most Active Thread",
                f"{most_active['name']} ({most_active['message_count']:,} msgs, {chat_type})",
            )

        most_active_group = conversations.get("most_active_group_chat")
        if most_active_group and most_active_group.get("name"):
            table.add_row(
                "Most Active Group",
                f"{most_active_group['name']} ({most_active_group['message_count']:,} msgs)",
            )

        self.console.print(table)

    def _render_response_times_section(self, response_times: dict[str, Any]) -> None:
        if (
            response_times.get("total_responses_you", 0) == 0
            and response_times.get("total_responses_them", 0) == 0
        ):
            return

        self.console.print("\n[bold cyan]⚡ Response Times[/]")

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("Metric", style="dim")
        table.add_column("Value", style="green")

        if response_times.get("total_responses_you", 0) > 0:
            table.add_row(
                "Your Median Response Time",
                response_times.get("median_response_time_you_formatted", "N/A"),
            )

        if response_times.get("total_responses_them", 0) > 0:
            table.add_row(
                "Their Median Response Time",
                response_times.get("median_response_time_them_formatted", "N/A"),
            )

        self.console.print(table)

    def _render_tapbacks_section(self, tapbacks: dict[str, Any]) -> None:
        if (
            tapbacks.get("total_tapbacks_given", 0) == 0
            and tapbacks.get("total_tapbacks_received", 0) == 0
        ):
            return

        self.console.print("\n[bold cyan]❤️ Reactions & Tapbacks[/]")

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("Metric", style="dim")
        table.add_column("Value", style="red")

        table.add_row("Total Tapbacks Given", f"{tapbacks.get('total_tapbacks_given', 0):,}")
        table.add_row("Total Tapbacks Received", f"{tapbacks.get('total_tapbacks_received', 0):,}")

        fav = tapbacks.get("favorite_tapback", (None, 0))
        if fav[0]:
            table.add_row("Your Favorite Reaction", f"{fav[0]} ({fav[1]:,} times)")

        most_received = tapbacks.get("most_received_tapback", (None, 0))
        if most_received[0]:
            table.add_row(
                "Most Received Reaction", f"{most_received[0]} ({most_received[1]:,} times)"
            )

        self.console.print(table)

    def _render_ghosts_section(self, ghosts: dict[str, Any]) -> None:
        if not ghosts:
            return

        total_you = ghosts.get("people_you_left_hanging", 0)
        total_them = ghosts.get("people_who_left_you_hanging", 0)
        if total_you == 0 and total_them == 0:
            return

        self.console.print("\n[bold cyan]👻 Ghost Mode[/]")

        table = Table(show_header=False, box=None, padding=(0, 2))
        table.add_column("Metric", style="dim")
        table.add_column("Value", style="magenta")

        timeline = ghosts.get("timeline_days")
        min_consecutive = ghosts.get("min_consecutive_messages")
        min_conversation = ghosts.get("min_conversation_messages")

        if timeline:
            table.add_row("Silence Threshold", f"{timeline} days without a reply")
        if min_consecutive:
            table.add_row("Minimum Texts in a Row", f"{min_consecutive} messages")
        if min_conversation:
            table.add_row("Minimum Messages per Chat", f"{min_conversation}")

        table.add_row("People You Left Hanging", f"{total_you:,}")
        table.add_row("People Who Left You Hanging", f"{total_them:,}")

        ratio = ghosts.get("ghost_ratio")
        if ratio is not None:
            table.add_row("Ghost Ratio (You/Them)", f"{ratio:.2f}")

        self.console.print(table)

    def _render_cliffhangers_section(self, cliffhangers: dict[str, Any] | None) -> None:
        if not cliffhangers:
            return

        count_you = cliffhangers.get("count", 0)
        count_them = cliffhangers.get("count_them", 0)
        if count_you <= 0 and count_them <= 0:
            return

        threshold = cliffhangers.get("threshold_hours", 12)
        longest_wait_you = cliffhangers.get("longest_wait_hours")
        longest_wait_them = cliffhangers.get("longest_wait_hours_them")
        self.console.print("\n[bold cyan]🧵 Cliffhangers[/]")
        if count_you > 0:
            self.console.print(
                f"You dangled future gossip [bold]{count_you}[/] times (took ≥{threshold}h to follow up)."
            )
            if isinstance(longest_wait_you, (int, float)) and longest_wait_you > 0:
                self.console.print(
                    f"[dim]Longest you made someone wait: {longest_wait_you:.1f}h[/]"
                )
        if count_them > 0:
            self.console.print(
                f"Your friends dangled it back [bold]{count_them}[/] times (you waited ≥{threshold}h)."
            )
            if isinstance(longest_wait_them, (int, float)) and longest_wait_them > 0:
                self.console.print(f"[dim]Longest they made you wait: {longest_wait_them:.1f}h[/]")

        examples_you = cliffhangers.get("examples") or []
        if examples_you:
            self.console.print("[dim]Your slowest follow-ups:[/]")
            for example in examples_you:
                contact = example.get("contact") or "Unknown"
                snippet = (example.get("snippet") or "").strip()
                when = example.get("timestamp") or "unknown date"
                wait_hours = example.get("hours_waited")
                preview = snippet if len(snippet) <= 60 else f"{snippet[:57]}..."
                details: list[str] = []
                if isinstance(wait_hours, (int, float)):
                    details.append(f"{wait_hours:.1f}h wait")
                if when and when != "unknown date":
                    details.append(when)
                meta = " • ".join(details) if details else "unknown date"
                self.console.print(f" • {contact}: “{preview}” ({meta})")

        examples_them = cliffhangers.get("examples_them") or []
        if examples_them:
            self.console.print("[dim]Times they left you hanging:[/]")
            for example in examples_them:
                contact = example.get("contact") or "Unknown"
                snippet = (example.get("snippet") or "").strip()
                when = example.get("timestamp") or "unknown date"
                wait_hours = example.get("hours_waited")
                preview = snippet if len(snippet) <= 60 else f"{snippet[:57]}..."
                details: list[str] = []
                if isinstance(wait_hours, (int, float)):
                    details.append(f"{wait_hours:.1f}h wait")
                if when and when != "unknown date":
                    details.append(when)
                meta = " • ".join(details) if details else "unknown date"
                self.console.print(f" • {contact}: “{preview}” ({meta})")

    def _render_stub_section(self, title: str, stub_data: dict[str, Any]) -> None:
        self.console.print(f"\n[bold cyan]{title}[/]")

        if stub_data.get("status") == "not_implemented":
            self.console.print(f"[dim]{stub_data.get('message', 'Not yet implemented')}[/]")

    def _sentiment_percentages(self, distribution: dict[str, int]) -> dict[str, str]:
        total = sum(distribution.values())
        if total == 0:
            return {"positive": "0%", "neutral": "0%", "negative": "0%"}

        def _format(value: int) -> str:
            percent = (value / total) * 100
            return f"{percent:.0f}%"

        return {
            "positive": _format(distribution.get("positive", 0)),
            "neutral": _format(distribution.get("neutral", 0)),
            "negative": _format(distribution.get("negative", 0)),
        }

    def _format_period_label(self, period: str, interval: str) -> str:
        if interval == "month":
            try:
                dt = datetime.strptime(period, "%Y-%m")
                return dt.strftime("%b %Y")
            except ValueError:
                return period
        if interval == "week":
            return period.replace("-", " ")
        return period

    def _format_sentiment_value(self, entry: dict[str, Any] | None) -> str:
        if not entry:
            return "—"
        avg = entry.get("avg_score")
        if avg is None:
            return "—"
        return f"{avg:+.2f}"

    def _merge_period_rows(
        self,
        sent: list[dict[str, Any]],
        received: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        sent_map = {entry["period"]: entry for entry in sent}
        received_map = {entry["period"]: entry for entry in received}
        periods = sorted(set(sent_map) | set(received_map))
        rows = []
        for period in periods:
            rows.append(
                {
                    "period": period,
                    "sent": sent_map.get(period),
                    "received": received_map.get(period),
                }
            )
        return rows



from __future__ import annotations

import math
import re
from dataclasses import dataclass
from typing import Mapping, Sequence

from .lexicon import BOOSTER_WORDS, LEXICON, NEGATIONS

__all__ = ["SentimentResult", "LexicalSentimentAnalyzer"]


EMOJI_SENTIMENT = {
    "❤️": 2.5,
    "💖": 2.2,
    "😍": 2.2,
    "🥰": 2.0,
    "😘": 1.8,
    "😊": 1.5,
    "🙂": 1.0,
    "🤗": 1.5,
    "🔥": 1.2,
    "👍": 1.0,
    "👏": 1.2,
    "😂": 1.1,
    "🤣": 1.1,
    "😁": 1.0,
    "😎": 0.9,
    "🙌": 1.0,
    "💯": 1.4,
    "😢": -1.4,
    "😭": -2.2,
    "😞": -1.2,
    "😔": -1.0,
    "😠": -1.5,
    "😡": -1.8,
    "🤬": -2.0,
    "💔": -1.8,
    "👎": -1.2,
    "🙁": -0.9,
    "😱": -1.6,
    "🤢": -1.7,
    "🤮": -2.2,
    "🥴": -0.8,
    "😤": -1.1,
    "😩": -1.4,
}

EMOTICON_SENTIMENT = {
    ":)": 1.2,
    "(:": 1.2,
    ":D": 1.6,
    "xD": 1.4,
    "XD": 1.4,
    ";)": 0.8,
    "<3": 1.8,
    ":-)": 1.2,
    "=)": 1.0,
    ":(": -1.4,
    "):": -1.4,
    ":'(": -1.8,
    ":-(": -1.4,
    "=(": -1.4,
    ">:(": -1.8,
    ":-/": -0.8,
    ":-|": -0.4,
}

IDIOM_SENTIMENT: dict[Sequence[str], float] = {
    ("not", "bad"): 1.5,
    ("not", "good"): -1.4,
    ("kind", "of", "annoyed"): -1.2,
    ("kind", "of", "excited"): 1.2,
    ("sick", "of"): -1.6,
    ("sick", "in", "a", "good", "way"): 1.7,
    ("no", "way"): -1.1,
    ("love", "you"): 1.8,
    ("miss", "you"): -0.8,
    ("can't", "wait"): 1.6,
    ("so", "over", "it"): -1.5,
    ("lost", "for", "words"): -1.1,
    ("over", "the", "moon"): 1.9,
    ("happy", "for", "you"): 1.4,
}

CASUAL_INTENSIFIERS = {
    "super": 0.65,
    "mega": 0.7,
    "hella": 0.8,
    "crazy": 0.5,
    "mad": 0.45,
}


@dataclass(frozen=True)
class SentimentResult:
    score: float  # normalized to [-1, 1]
    label: str  # "positive", "neutral", or "negative"


class LexicalSentimentAnalyzer:
    """Simple rule-based sentiment scorer with a tiny footprint."""

    WORD_PATTERN = re.compile(r"[A-Za-z0-9#'@]+")
    SENTIMENT_THRESHOLD = 0.12
    NORMALIZER = 1.5
    MAX_EXCLAMATION_EMPHASIS = 4
    NEGATION_WINDOW = 2
    MAX_TOKEN_REDUCTION = 4
    VARIANCE_BASELINE = 0.15

    def __init__(self, lexicon: Mapping[str, float] | None = None):
        self._lexicon = {k: float(v) for k, v in (lexicon or LEXICON).items()}

    def analyze(self, text: str | None) -> SentimentResult:
        if not text:
            return SentimentResult(0.0, "neutral")

        tokens = [self._normalize_token(tok) for tok in self.WORD_PATTERN.findall(text)]
        tokens = [tok for tok in tokens if tok]
        if not tokens:
            return self._emoji_only_score(text)

        tokens, idiom_bonus = self._apply_idiom_scores(tokens)

        total_score = 0.0
        hit_count = 0
        negate_window = 0
        previous_token: str | None = None

        for token in tokens:
            if token in NEGATIONS:
                negate_window = self.NEGATION_WINDOW
                previous_token = token
                continue

            lex_score = self._lexicon.get(token)
            modifier = 1.0

            if previous_token and previous_token in BOOSTER_WORDS:
                modifier += BOOSTER_WORDS[previous_token]
            elif previous_token and previous_token in CASUAL_INTENSIFIERS:
                modifier += CASUAL_INTENSIFIERS[previous_token]

            if lex_score is None and token.startswith("#"):
                lex_score = self._lexicon.get(token[1:])

            if lex_score is None:
                previous_token = token
                if negate_window:
                    negate_window -= 1
                continue

            if negate_window:
                modifier *= -0.75
                negate_window -= 1

            if token.isupper() and len(token) > 1:
                modifier += 0.15

            total_score += lex_score * modifier
            hit_count += 1
            previous_token = token

        if hit_count == 0:
            emoji_result = self._emoji_only_score(text)
            if emoji_result.label != "neutral":
                return emoji_result
            return SentimentResult(0.0, "neutral")

        total_score += idiom_bonus
        total_score += self._emoji_bonus(text)
        total_score += self._emoticon_bonus(text)

        normalized = total_score / (hit_count * self.NORMALIZER)
        normalized = max(min(normalized, 1.0), -1.0)

        exclamations = min(text.count("!"), self.MAX_EXCLAMATION_EMPHASIS)
        if exclamations and normalized != 0:
            normalized += math.copysign(exclamations * 0.02, normalized)

        if "??" in text:
            normalized -= 0.05

        normalized = max(min(normalized, 1.0), -1.0)
        normalized = self._boost_variance(normalized)
        label = self._label_from_score(normalized)
        return SentimentResult(round(normalized, 3), label)

    def _label_from_score(self, score: float) -> str:
        if score >= self.SENTIMENT_THRESHOLD:
            return "positive"
        if score <= -self.SENTIMENT_THRESHOLD:
            return "negative"
        return "neutral"

    def _normalize_token(self, token: str) -> str:
        lowered = token.lower()
        if lowered.startswith("@"):
            return ""
        if lowered.startswith("#") and len(lowered) > 1:
            lowered = "#" + lowered[1:]
        lowered = self._squash_repeated_chars(lowered)
        return lowered

    def _squash_repeated_chars(self, token: str) -> str:
        if len(token) <= 3:
            return token
        result = []
        last_char = ""
        repeat_count = 0
        for char in token:
            if char == last_char:
                repeat_count += 1
            else:
                repeat_count = 1
                last_char = char
            if repeat_count <= self.MAX_TOKEN_REDUCTION:
                result.append(char)
        return "".join(result)

    def _apply_idiom_scores(self, tokens: list[str]) -> tuple[list[str], float]:
        if not tokens:
            return tokens, 0.0

        adjustments = 0.0
        consumed: set[int] = set()
        for start in range(len(tokens)):
            if start in consumed:
                continue
            for phrase, score in IDIOM_SENTIMENT.items():
                length = len(phrase)
                if length == 0 or start + length > len(tokens):
                    continue
                if tokens[start : start + length] == list(phrase):
                    adjustments += score
                    for idx in range(start, start + length):
                        consumed.add(idx)
                    break

        filtered = [tok for idx, tok in enumerate(tokens) if idx not in consumed]
        return filtered, adjustments

    def _emoji_bonus(self, text: str) -> float:
        bonus = 0.0
        for char in text:
            if char in EMOJI_SENTIMENT:
                bonus += EMOJI_SENTIMENT[char]
        return bonus * 0.2

    def _emoticon_bonus(self, text: str) -> float:
        bonus = 0.0
        for emoticon, value in EMOTICON_SENTIMENT.items():
            if emoticon in text:
                bonus += value
        return bonus * 0.3

    def _emoji_only_score(self, text: str) -> SentimentResult:
        emoji_score = self._emoji_bonus(text) + self._emoticon_bonus(text)
        if emoji_score == 0:
            return SentimentResult(0.0, "neutral")
        normalized = max(min(emoji_score / self.NORMALIZER, 1.0), -1.0)
        normalized = self._boost_variance(normalized)
        return SentimentResult(
            round(normalized, 3),
            self._label_from_score(normalized),
        )

    def _boost_variance(self, score: float) -> float:
        """Push magnitudes away from zero while preserving sign."""
        magnitude = abs(score)
        if magnitude == 0.0:
            return 0.0
        baseline = max(self.VARIANCE_BASELINE, 1e-6)
        stretched = 1 - math.exp(-(magnitude / baseline))
        return math.copysign(min(stretched, 1.0), score)



"""
Phrase analysis utilities for surfacing most-used expressions.

This package intentionally mirrors the sentiment module layout: the public API
exposes lightweight dataclasses and a single `PhraseExtractor` facade while the
implementation details (tokenization heuristics, TF-IDF scoring, stopwords,
etc.) live in dedicated submodules.  That makes it easy to plug the extractor
into the analyzer/CLI later without dragging extra dependencies into callers.
"""

from .extractor import (
    ContactPhraseStats,
    PhraseExtractionConfig,
    PhraseExtractionResult,
    PhraseExtractor,
    PhraseStat,
)

__all__ = [
    "PhraseExtractionConfig",
    "PhraseExtractionResult",
    "PhraseExtractor",
    "PhraseStat",
    "ContactPhraseStats",
]



import os
import sqlite3

from rich.console import Console
from rich.panel import Panel

DEFAULT_DB_PATH = "~/Library/Messages/chat.db"


class PermissionError(Exception):
    pass


def check_database_access(db_path: str | None = None) -> bool:
    path = db_path or os.path.expanduser(DEFAULT_DB_PATH)

    try:
        conn = sqlite3.connect(f"file:{path}?mode=ro", uri=True)
        conn.execute("SELECT 1 FROM message LIMIT 1")
        conn.close()
        return True
    except (sqlite3.OperationalError, sqlite3.DatabaseError):
        return False


def require_database_access(db_path: str | None = None) -> None:
    if not check_database_access(db_path):
        display_permission_error()
        raise PermissionError("Full Disk Access required to read iMessage database")


def display_permission_error() -> None:
    console = Console()
    console.print(
        Panel(
            "[bold red]Full Disk Access Required[/]\n\n"
            "This application needs permission to read your iMessage database.\n\n"
            "[bold]To grant access:[/]\n"
            "1. Open [cyan]System Settings[/]\n"
            "2. Go to [cyan]Privacy & Security → Full Disk Access[/]\n"
            "3. Click [cyan]+[/] and add this application or Terminal\n"
            "4. Restart the application",
            title="⚠️  Permission Error",
        )
    )



import argparse
import logging
import sys
from pathlib import Path

import questionary
from rich.console import Console
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TaskID,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

from . import (
    Exporter,
    ExportLoader,
    MessageService,
    NLPStatisticsAnalyzer,
    PermissionError,
    RawStatisticsAnalyzer,
    TerminalDisplay,
    require_database_access,
)
from .phrase_utils import compute_phrases_for_export
from .sentiment_utils import compute_sentiment_for_export
from .utils import sanitize_statistics_for_export

logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser(
        prog="imexport",
        description="Export and analyze iMessage conversations from macOS",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "-y",
        "--year",
        type=int,
        default=2025,
        help="Year to export (default: 2025)",
    )

    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Output file path (default: exports/imessage_export_YEAR.jsonl)",
    )

    parser.add_argument(
        "-d",
        "--database",
        type=str,
        help="Path to chat.db (default: ~/Library/Messages/chat.db)",
    )

    parser.add_argument(
        "--format",
        type=str,
        choices=["jsonl", "json"],
        default="jsonl",
        help="Export format (default: jsonl)",
    )

    parser.add_argument(
        "--indent",
        type=int,
        default=2,
        help="JSON indentation spaces for json format (default: 2, use 0 for compact)",
    )

    parser.add_argument(
        "--skip-permission-check",
        action="store_true",
        help="Skip permission check (use with caution)",
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug logging",
    )

    parser.add_argument(
        "--replace-cache",
        action="store_true",
        help="Replace existing cached export file if it exists",
    )

    parser.add_argument(
        "--no-analyze",
        action="store_true",
        help="Skip analysis after export",
    )

    parser.add_argument(
        "input",
        type=str,
        nargs="?",
        help="Path to exported JSON/JSONL file (for analyze-only mode)",
    )

    parser.add_argument(
        "--analyzers",
        type=str,
        default="raw",
        help="Comma-separated list of analyzers to run (raw,nlp) (default: raw)",
    )

    parser.add_argument(
        "--stats-output",
        type=str,
        dest="stats_output",
        help="Output file path for statistics JSON (optional)",
    )

    parser.add_argument(
        "--no-share",
        action="store_false",
        dest="share",
        help="Don't upload statistics (show full terminal output instead)",
    )

    parser.add_argument(
        "--share",
        action="store_true",
        dest="share",
        default=True,
        help="Upload statistics to web and get shareable link (default)",
    )

    parser.add_argument(
        "--server-url",
        type=str,
        default="https://imessage-wrapped.fly.dev",
        help="Web server URL for sharing (default: https://imessage-wrapped.fly.dev)",
    )

    parser.add_argument(
        "--dev",
        action="store_true",
        help="Use local development server (http://localhost:3000)",
    )

    parser.add_argument(
        "--ghost-timeline",
        type=int,
        default=7,
        help="Days without a reply before someone counts as a ghost (default: 7)",
    )

    parser.add_argument(
        "--compare",
        nargs=2,
        type=int,
        metavar=("YEAR1", "YEAR2"),
        help="Compare two years (e.g., --compare 2024 2025). Creates both individual wraps and a comparison view.",
    )

    args = parser.parse_args()

    if args.dev:
        args.replace_cache = True

    return args


def export_command(args):
    console = Console()

    if not args.skip_permission_check:
        try:
            require_database_access(args.database)
        except PermissionError:
            sys.exit(1)

    if args.output:
        output_path = args.output
    else:
        ext = "jsonl" if args.format == "jsonl" else "json"
        output_path = f"exports/imessage_export_{args.year}.{ext}"

    output_file = Path(output_path)

    if output_file.exists() and not args.replace_cache:
        console.print(f"\n[yellow]ℹ[/] Export file already exists: [cyan]{output_path}[/]")
        console.print("[dim]Use --replace-cache to regenerate[/]")
        return output_path

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,
    ) as progress:
        task = progress.add_task(f"Exporting messages from {args.year}...", total=None)

        service = MessageService(db_path=args.database)
        data = service.export_year(args.year)

        # Precompute phrases while text is available; stored alongside export without raw text.
        phrases, phrases_by_contact = compute_phrases_for_export(data)
        data.phrases = phrases or None
        # Per-contact phrases are intentionally omitted from export for privacy.
        data.phrases_by_contact = None

        # Precompute sentiment (overall + monthly) while text is available.
        data.sentiment = compute_sentiment_for_export(data) or None

        progress.update(task, description=f"Writing {data.total_messages} messages to file...")

        from .exporter import JSONLSerializer, JSONSerializer

        if args.format == "json":
            serializer = JSONSerializer(indent=args.indent if args.indent > 0 else None)
        else:
            serializer = JSONLSerializer()
        exporter = Exporter(serializer=serializer)
        exporter.export_to_file(data, output_path)

    console.print(
        f"\n[green]✓[/] Exported {data.total_messages} messages to [cyan]{output_path}[/]"
    )
    console.print(f"[dim]Conversations: {len(data.conversations)}[/]")

    return output_path


def analyze_command(args, input_path=None):
    console = Console()

    if args.ghost_timeline <= 0:
        console.print("[red]✗[/] --ghost-timeline must be greater than zero")
        sys.exit(1)

    if input_path:
        input_path = Path(input_path)
        if not input_path.exists():
            console.print(f"[red]✗[/] Input file not found: {input_path}")
            sys.exit(1)
    elif args.input:
        input_path = Path(args.input)
        if not input_path.exists():
            console.print(f"[red]✗[/] Input file not found: {args.input}")
            sys.exit(1)
    else:
        exports_dir = Path("exports")
        export_files = []

        if exports_dir.exists():
            export_files = sorted(
                [f for f in exports_dir.iterdir() if f.suffix in [".json", ".jsonl"]],
                key=lambda x: x.stat().st_mtime,
                reverse=True,
            )

        if not export_files:
            console.print("[yellow]ℹ[/] No export found. Exporting messages first...\n")

            export_args = argparse.Namespace(
                year=args.year,
                output=None,
                database=args.database,
                format="jsonl",
                indent=2,
                skip_permission_check=args.skip_permission_check,
                debug=args.debug,
                replace_cache=args.replace_cache,
            )
            export_command(export_args)

            if not exports_dir.exists():
                console.print("[red]✗[/] Export failed.")
                sys.exit(1)

            export_files = sorted(
                [f for f in exports_dir.iterdir() if f.suffix in [".json", ".jsonl"]],
                key=lambda x: x.stat().st_mtime,
                reverse=True,
            )

        if not export_files:
            console.print("[red]✗[/] Export failed.")
            sys.exit(1)

        if args.share or len(export_files) == 1:
            input_path = export_files[0]
        else:
            choices = []
            for file in export_files:
                size_mb = file.stat().st_size / (1024 * 1024)
                choices.append(
                    questionary.Choice(title=f"{file.name} ({size_mb:.1f} MB)", value=file)
                )

            selected = questionary.select("Select export file to analyze:", choices=choices).ask()

            if selected is None:
                sys.exit(0)

            input_path = selected

    analyzer_names = [name.strip() for name in args.analyzers.split(",")]

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=None, complete_style="green"),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
        TimeRemainingColumn(elapsed_when_finished=True),
        console=console,
        transient=True,
    ) as progress:
        load_task = progress.add_task("Loading export data...", total=1)

        try:
            data = ExportLoader.load(input_path)
        except Exception as e:
            console.print(f"[red]✗[/] Failed to load export data: {e}")
            sys.exit(1)

        progress.update(load_task, advance=1, description="Export loaded")

        sentiment_tasks: dict[str, TaskID] = {}

        def sentiment_progress(stage: str, completed: int, total: int) -> None:
            if total == 0:
                return
            label = "Sentiment (You)" if stage == "sent" else "Sentiment (Them)"
            task_id = sentiment_tasks.get(stage)
            if task_id is None:
                task_id = progress.add_task(label, total=total)
                sentiment_tasks[stage] = task_id
            progress.update(task_id, completed=completed, total=total)

        analyzers = []

        def _print_sentiment_info(info: dict[str, str | int | None] | None) -> None:
            if not info:
                return
            params = info.get("parameters_display")
            name = info.get("name") or "Sentiment"
            details: list[str] = []
            if params:
                details.append(f"{params} parameters")
            detail_text = f" ({', '.join(details)})" if details else ""
            progress.console.print(f"[dim]Using sentiment backend: {name}{detail_text}[/]")

        if "raw" in analyzer_names:
            analyzers.append(
                RawStatisticsAnalyzer(
                    sentiment_progress=sentiment_progress,
                    ghost_timeline_days=args.ghost_timeline,
                )
            )
            _print_sentiment_info(analyzers[-1].sentiment_model_info)
        if "nlp" in analyzer_names:
            analyzers.append(NLPStatisticsAnalyzer())
        analyzer_task = progress.add_task(
            f"Running {len(analyzers)} analyzer(s)...", total=max(len(analyzers), 1)
        )

        statistics = {}
        for analyzer in analyzers:
            progress.update(analyzer_task, description=f"Running {analyzer.name} analyzer...")
            statistics[analyzer.name] = analyzer.analyze(data)
            progress.advance(analyzer_task)

    sanitized_statistics = sanitize_statistics_for_export(statistics)

    if args.stats_output:
        import json

        output_path = Path(args.stats_output)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        with open(output_path, "w", encoding="utf-8") as f:
            json.dump(sanitized_statistics, f, indent=2, ensure_ascii=False)
        console.print(f"\n[green]✓[/] Statistics saved to [cyan]{args.stats_output}[/]")

    display = TerminalDisplay()
    display.render(statistics, brief=args.share)

    if args.share:
        from .uploader import StatsUploader

        server_url = "http://localhost:3000" if args.dev else args.server_url
        uploader = StatsUploader(base_url=server_url)

        year = data.year if hasattr(data, "year") else 2025
        user_name = data.user_name if hasattr(data, "user_name") else None
        share_url = uploader.upload(year, sanitized_statistics, user_name=user_name)

        if not share_url:
            console.print("\n[yellow]Tip: Make sure the web server is running:[/]")
            console.print("[dim]  cd web && npm install && npm run dev[/]")


def compare_command(args):
    """Handle year-over-year comparison"""
    console = Console()

    year1, year2 = args.compare

    # Ensure year1 < year2 for consistency
    if year1 > year2:
        year1, year2 = year2, year1

    if year1 == year2:
        console.print("[red]✗[/] Cannot compare the same year")
        sys.exit(1)

    console.print(f"\n[bold cyan]Creating year-over-year comparison: {year1} vs {year2}[/]\n")

    # Export and analyze year 1
    console.print(f"[cyan]━━━ Analyzing {year1} ━━━[/]")
    args_year1 = argparse.Namespace(**vars(args))
    args_year1.year = year1
    args_year1.share = False  # Don't upload individual years yet
    export_path1 = export_command(args_year1)

    if not export_path1:
        console.print(f"[red]✗[/] Failed to export {year1}")
        sys.exit(1)

    # Temporarily suppress sharing for individual analysis
    stats1 = {}
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,
        transient=True,
    ) as progress:
        load_task = progress.add_task(f"Loading {year1} export data...", total=1)
        data1 = ExportLoader.load(export_path1)
        progress.update(load_task, advance=1)

        analyzer = RawStatisticsAnalyzer(ghost_timeline_days=args.ghost_timeline)
        stats1 = analyzer.analyze(data1)

    console.print(f"[green]✓[/] {year1} analysis complete\n")

    # Export and analyze year 2
    console.print(f"[cyan]━━━ Analyzing {year2} ━━━[/]")
    args_year2 = argparse.Namespace(**vars(args))
    args_year2.year = year2
    args_year2.share = False
    export_path2 = export_command(args_year2)

    if not export_path2:
        console.print(f"[red]✗[/] Failed to export {year2}")
        sys.exit(1)

    stats2 = {}
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,
        transient=True,
    ) as progress:
        load_task = progress.add_task(f"Loading {year2} export data...", total=1)
        data2 = ExportLoader.load(export_path2)
        progress.update(load_task, advance=1)

        analyzer = RawStatisticsAnalyzer(ghost_timeline_days=args.ghost_timeline)
        stats2 = analyzer.analyze(data2)

    console.print(f"[green]✓[/] {year2} analysis complete\n")

    # Sanitize statistics for both years
    from .utils import sanitize_statistics_for_export

    sanitized_stats1 = sanitize_statistics_for_export({"raw": stats1})
    sanitized_stats2 = sanitize_statistics_for_export({"raw": stats2})

    # Upload comparison
    console.print("[cyan]━━━ Creating comparison view ━━━[/]")
    from .uploader import ComparisonUploader

    server_url = "http://localhost:3000" if args.dev else args.server_url
    uploader = ComparisonUploader(base_url=server_url)

    user_name1 = data1.user_name if hasattr(data1, "user_name") else None
    user_name2 = data2.user_name if hasattr(data2, "user_name") else None
    user_name = user_name1 or user_name2  # Use whichever is available

    comparison_url = uploader.upload_comparison(
        year1, year2, sanitized_stats1, sanitized_stats2, user_name=user_name
    )

    if not comparison_url:
        console.print("\n[yellow]Tip: Make sure the web server is running:[/]")
        console.print("[dim]  cd web && npm install && npm run dev[/]")


def main():
    args = parse_args()

    if args.debug:
        logging.basicConfig(
            level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        logger.debug("Debug logging enabled")

    # Handle comparison mode
    if args.compare:
        compare_command(args)
        return

    if args.input:
        analyze_command(args)
    else:
        export_path = export_command(args)
        if export_path and not args.no_analyze:
            analyze_command(args, input_path=export_path)


if __name__ == "__main__":
    main()



"""Collect system and SDK metadata for uploads."""

import platform
import sys
from typing import Optional


def get_sdk_version() -> str:
    """Get the Python SDK version."""
    try:
        from importlib.metadata import version

        return version("imessage-wrapped")
    except Exception:
        # Fallback to reading from pyproject.toml if importlib fails
        try:
            import re
            from pathlib import Path

            # Try to find pyproject.toml
            current_file = Path(__file__)
            for parent in [current_file.parent.parent.parent, current_file.parent.parent]:
                pyproject = parent / "pyproject.toml"
                if pyproject.exists():
                    content = pyproject.read_text()
                    match = re.search(r'version\s*=\s*"([^"]+)"', content)
                    if match:
                        return match.group(1)
        except Exception:
            pass
        return "unknown"


def get_dmg_version() -> Optional[str]:
    """
    Get the DMG version if running from a packaged app.
    Returns None if not applicable or cannot be determined.
    """
    try:
        # Check if running from a .app bundle
        if sys.executable and ".app" in sys.executable.lower():
            # Try to read version from app bundle
            from pathlib import Path

            app_path = Path(sys.executable)

            # Look for Info.plist in the app bundle
            for parent in app_path.parents:
                if parent.suffix == ".app":
                    info_plist = parent / "Contents" / "Info.plist"
                    if info_plist.exists():
                        # Simple regex extraction (avoiding plistlib for simplicity)
                        import re

                        content = info_plist.read_text(errors="ignore")
                        match = re.search(
                            r"<key>CFBundleShortVersionString</key>\s*<string>([^<]+)</string>",
                            content,
                        )
                        if match:
                            return match.group(1)
                    break
    except Exception:
        pass
    return None


def collect_metadata() -> dict:
    """
    Collect system and SDK metadata.
    This should never raise an exception.
    """
    metadata = {}

    try:
        metadata["sdk_version"] = get_sdk_version()
    except Exception:
        metadata["sdk_version"] = "unknown"

    try:
        dmg_version = get_dmg_version()
        if dmg_version:
            metadata["dmg_version"] = dmg_version
    except Exception:
        pass

    try:
        metadata["python_version"] = (
            f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        )
    except Exception:
        pass

    try:
        metadata["platform"] = platform.system()
    except Exception:
        pass

    try:
        metadata["platform_version"] = platform.release()
    except Exception:
        pass

    try:
        metadata["machine"] = platform.machine()
    except Exception:
        pass

    return metadata



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from functools import lru_cache
from importlib import resources
from typing import Iterable, Mapping, Sequence

from .scoring import FrequencyScorer, ScoredPhrase, TfIdfScorer
from .tokenizer import SimpleTokenizer, TokenizedMessage

STOPWORDS_PACKAGE = "imessage_wrapped.phrases.resources"
STOPWORDS_FILE = "stopwords_en.txt"

__all__ = [
    "PhraseExtractionConfig",
    "PhraseExtractionResult",
    "PhraseExtractor",
    "PhraseStat",
    "ContactPhraseStats",
]


@dataclass(frozen=True)
class PhraseStat:
    text: str
    occurrences: int
    share: float


@dataclass(frozen=True)
class ContactPhraseStats:
    contact_id: str
    contact_name: str | None
    total_messages: int
    top_phrases: list[PhraseStat] = field(default_factory=list)


@dataclass(frozen=True)
class PhraseExtractionResult:
    overall: list[PhraseStat]
    by_contact: list[ContactPhraseStats]
    analyzed_messages: int
    config: "PhraseExtractionConfig"


@dataclass(frozen=True)
class PhraseExtractionConfig:
    ngram_range: tuple[int, int] = (2, 5)
    min_occurrences: int = 3
    min_characters: int = 4
    min_text_messages: int = 50
    per_contact_min_text_messages: int | None = 20
    max_phrases: int = 10  # number of phrases to return
    per_contact_limit: int = 500
    scoring: str = "tfidf"  # or "tfidf" / "frequency"
    dedupe_overlap: bool = True
    overlap_tolerance: float = 0.1
    contains_letters: bool = True
    length_bias: float = 2.1
    phrase_filter_bank: list[str] = field(default_factory=lambda: ["http", "https"])

    def __post_init__(self) -> None:
        if self.ngram_range[0] <= 0 or self.ngram_range[0] > self.ngram_range[1]:
            raise ValueError("ngram_range must be a tuple like (1, 3)")
        if self.min_occurrences < 1:
            raise ValueError("min_occurrences must be >= 1")
        if self.min_characters < 1:
            raise ValueError("min_characters must be >= 1")
        if self.min_text_messages < 1:
            raise ValueError("min_text_messages must be >= 1")
        if (
            self.per_contact_min_text_messages is not None
            and self.per_contact_min_text_messages < 1
        ):
            raise ValueError("per_contact_min_text_messages must be >= 1")
        if self.max_phrases < 1:
            raise ValueError("max_phrases must be >= 1")
        if self.scoring not in {"frequency", "tfidf"}:
            raise ValueError("scoring must be 'frequency' or 'tfidf'")
        if self.per_contact_limit < 0:
            raise ValueError("per_contact_limit must be >= 0")
        if self.length_bias < 0:
            raise ValueError("length_bias must be >= 0")
        if not isinstance(self.contains_letters, bool):
            raise ValueError("contains_letters must be a boolean")
        if self.length_bias < 0:
            raise ValueError("length_bias must be >= 0")
        if not isinstance(self.phrase_filter_bank, list):
            raise ValueError("phrase_filter_bank must be a list of strings")
        if not all(isinstance(item, str) for item in self.phrase_filter_bank):
            raise ValueError("phrase_filter_bank must be a list of strings")


class PhraseExtractor:
    def __init__(
        self,
        config: PhraseExtractionConfig | None = None,
        stopwords: Iterable[str] | None = None,
        tokenizer: SimpleTokenizer | None = None,
    ) -> None:
        self._config = config or PhraseExtractionConfig()
        self._tokenizer = tokenizer or SimpleTokenizer()
        self._stopwords = self._build_stopword_set(stopwords)
        self._scorer = self._build_scorer(self._config.scoring)

    @property
    def config(self) -> PhraseExtractionConfig:
        return self._config

    def extract(
        self,
        messages: Sequence[str],
        *,
        per_contact_messages: Mapping[str, Sequence[str]] | None = None,
        contact_names: Mapping[str, str] | None = None,
    ) -> PhraseExtractionResult:
        overall_stats = self._extract_bucket(messages, min_messages=self._config.min_text_messages)

        contact_results: list[ContactPhraseStats] = []
        if per_contact_messages:
            for contact_id, contact_msgs in per_contact_messages.items():
                bucket = self._extract_bucket(
                    contact_msgs,
                    min_messages=(
                        self._config.per_contact_min_text_messages or self._config.min_text_messages
                    ),
                )
                contact_results.append(
                    ContactPhraseStats(
                        contact_id=contact_id,
                        contact_name=(contact_names or {}).get(contact_id),
                        total_messages=bucket.total_messages,
                        top_phrases=self._contact_phrase_slice(bucket.top_phrases),
                    )
                )

            contact_results.sort(
                key=lambda item: (
                    item.top_phrases[0].occurrences if item.top_phrases else 0,
                    item.total_messages,
                ),
                reverse=True,
            )

        return PhraseExtractionResult(
            overall=overall_stats.top_phrases[: self._config.max_phrases],
            by_contact=contact_results,
            analyzed_messages=overall_stats.total_messages,
            config=self._config,
        )

    def _extract_bucket(
        self,
        messages: Sequence[str],
        *,
        min_messages: int | None = None,
    ) -> "PhraseBucket":
        tokenized = self._tokenizer.tokenize_messages(messages)
        min_required = min_messages if min_messages is not None else self._config.min_text_messages
        if len(tokenized) < min_required:
            return PhraseBucket(total_messages=len(tokenized), top_phrases=[])

        counts, doc_frequencies, total_docs = self._count_phrases(tokenized)
        filtered_counts = {
            phrase: count
            for phrase, count in counts.items()
            if count >= self._config.min_occurrences
        }
        if not filtered_counts:
            return PhraseBucket(total_messages=len(tokenized), top_phrases=[])

        total_occurrences = sum(filtered_counts.values())
        scored = self._scorer.score(filtered_counts, doc_frequencies, total_docs)
        ranked = sorted(
            scored,
            key=lambda item: (
                item.score * self._length_multiplier(item.text),
                item.occurrences,
                len(item.text),
                item.text,
            ),
            reverse=True,
        )

        if self._config.dedupe_overlap:
            ranked = self._dedupe_overlaps(ranked)

        phrase_stats = [
            PhraseStat(
                text=item.text,
                occurrences=item.occurrences,
                share=round(item.occurrences / max(1, total_occurrences), 4),
            )
            for item in ranked
        ]

        return PhraseBucket(total_messages=len(tokenized), top_phrases=phrase_stats)

    def _contact_phrase_slice(self, phrases: Sequence[PhraseStat]) -> list[PhraseStat]:
        if self._config.per_contact_limit <= 0:
            return list(phrases)
        return list(phrases[: self._config.per_contact_limit])

    def _count_phrases(
        self,
        tokenized_messages: Sequence[TokenizedMessage],
    ) -> tuple[Counter[str], dict[str, int], int]:
        counts: Counter[str] = Counter()
        doc_frequencies: dict[str, int] = {}

        for message in tokenized_messages:
            seen_in_message: set[str] = set()
            tokens = message.tokens
            for phrase in self._generate_phrases(tokens):
                counts[phrase] += 1
                if phrase not in seen_in_message:
                    doc_frequencies[phrase] = doc_frequencies.get(phrase, 0) + 1
                    seen_in_message.add(phrase)

        return counts, doc_frequencies, len(tokenized_messages)

    def _generate_phrases(self, tokens: Sequence[str]) -> Iterable[str]:
        min_n, max_n = self._config.ngram_range
        length = len(tokens)
        for n in range(min_n, max_n + 1):
            if length < n:
                continue
            for idx in range(length - n + 1):
                window = tokens[idx : idx + n]
                if not self._valid_ngram(window):
                    continue
                yield " ".join(window)

    def _valid_ngram(self, tokens: Sequence[str]) -> bool:
        if not tokens:
            return False
        characters = sum(len(token) for token in tokens)
        if characters < self._config.min_characters:
            return False
        if self._config.contains_letters:
            has_letter = any(any(char.isalpha() for char in token) for token in tokens)
            if not has_letter:
                return False
        if self._config.phrase_filter_bank:
            has_filter = any(token in self._config.phrase_filter_bank for token in tokens)
            if has_filter:
                return False
        if all(token in self._stopwords for token in tokens):
            return False
        return True

    def _dedupe_overlaps(self, scored: Sequence[ScoredPhrase]) -> list[ScoredPhrase]:
        filtered: list[ScoredPhrase] = []
        for candidate in scored:
            skip = False
            for kept in filtered:
                if self._phrases_overlap(kept.text, candidate.text):
                    count_delta = abs(kept.occurrences - candidate.occurrences)
                    tolerance = max(1, int(kept.occurrences * self._config.overlap_tolerance))
                    if len(candidate.text) <= len(kept.text) and count_delta <= tolerance:
                        skip = True
                        break
            if not skip:
                filtered.append(candidate)

        result: list[ScoredPhrase] = []
        for phrase in filtered:
            is_subset = False
            for other in filtered:
                if phrase != other and phrase.text != other.text:
                    if self._is_phrase_subset(phrase.text, other.text):
                        is_subset = True
                        break
            if not is_subset:
                result.append(phrase)
        return result

    @staticmethod
    def _is_phrase_subset(shorter: str, longer: str) -> bool:
        if len(shorter) >= len(longer):
            return False
        shorter_words = shorter.split()
        longer_words = longer.split()
        if len(shorter_words) > len(longer_words):
            return False
        for i in range(len(longer_words) - len(shorter_words) + 1):
            if longer_words[i : i + len(shorter_words)] == shorter_words:
                return True
        return False

    @staticmethod
    def _phrases_overlap(left: str, right: str) -> bool:
        return left in right or right in left

    def _build_stopword_set(self, stopwords: Iterable[str] | None) -> set[str]:
        if stopwords is not None:
            return {word.strip().lower() for word in stopwords if word.strip()}
        return _load_stopwords()

    def _build_scorer(self, scoring: str):
        if scoring == "tfidf":
            return TfIdfScorer()
        return FrequencyScorer()

    def _length_multiplier(self, phrase: str) -> float:
        token_count = max(1, phrase.count(" ") + 1)
        extra_tokens = max(0, token_count - 1)
        factor = max(1.0, self._config.length_bias)
        return float(factor**extra_tokens)


@dataclass
class PhraseBucket:
    total_messages: int
    top_phrases: list[PhraseStat]


@lru_cache(maxsize=1)
def _load_stopwords() -> set[str]:
    base = resources.files(STOPWORDS_PACKAGE)
    with resources.as_file(base / STOPWORDS_FILE) as path:
        content = path.read_text(encoding="utf-8")
    return {line.strip() for line in content.splitlines() if line.strip()}



"""
Helpers for assigning scores to extracted phrases.

We keep the logic small and dependency-free so the feature can run in the
desktop/CLI environments without pulling in scikit-learn.
"""

from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Mapping

__all__ = ["ScoredPhrase", "TfIdfScorer", "FrequencyScorer"]


@dataclass(frozen=True)
class ScoredPhrase:
    text: str
    occurrences: int
    score: float


class FrequencyScorer:
    """Default scorer: the score is simply the frequency count as a float."""

    def score(
        self,
        phrase_counts: Mapping[str, int],
        doc_frequencies: Mapping[str, int],
        total_documents: int,
    ) -> list[ScoredPhrase]:
        return [
            ScoredPhrase(text=phrase, occurrences=count, score=float(count))
            for phrase, count in phrase_counts.items()
        ]


class TfIdfScorer:
    """Tiny TF-IDF implementation tailored for already-counted phrases."""

    def __init__(self, smooth: bool = True):
        self._smooth = smooth

    def score(
        self,
        phrase_counts: Mapping[str, int],
        doc_frequencies: Mapping[str, int],
        total_documents: int,
    ) -> list[ScoredPhrase]:
        if total_documents <= 0:
            return []

        scores: list[ScoredPhrase] = []
        for phrase, count in phrase_counts.items():
            doc_freq = max(1, doc_frequencies.get(phrase, 0))
            idf = self._idf(total_documents, doc_freq)
            scores.append(
                ScoredPhrase(
                    text=phrase,
                    occurrences=count,
                    score=count * idf,
                )
            )
        return scores

    def _idf(self, total_docs: int, doc_freq: int) -> float:
        if self._smooth:
            # Smooth by adding 1 to numerator/denominator to avoid division by zero
            return math.log((1 + total_docs) / (1 + doc_freq)) + 1.0
        return math.log(total_docs / doc_freq)



"""
Lightweight tokenization helpers tuned for conversational English text.

We purposely avoid heavy NLP dependencies; this module only normalizes text,
splits sentences, and yields lowercase word tokens.
"""

from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Iterator, Sequence

__all__ = ["TokenizedMessage", "SimpleTokenizer"]


_WORD_PATTERN = re.compile(r"[A-Za-z0-9']+")
_SENTENCE_BOUNDARY = re.compile(r"[.!?]+")
_APOSTROPHE_VARIANTS = {
    "\u2018": "'",
    "\u2019": "'",
    "\u2032": "'",
    "\u02bc": "'",
}


@dataclass(frozen=True)
class TokenizedMessage:
    tokens: list[str]


class SimpleTokenizer:
    """
    Minimal tokenizer for conversational text.

    * Normalizes curly apostrophes/quotes
    * Lowercases tokens
    * Splits on ASCII punctuation boundaries
    """

    def tokenize_messages(self, messages: Sequence[str]) -> list[TokenizedMessage]:
        return [TokenizedMessage(tokens=self.tokenize(text)) for text in messages if text]

    def tokenize(self, text: str) -> list[str]:
        normalized = self._normalize(text)
        return [match.group(0).lower() for match in _WORD_PATTERN.finditer(normalized)]

    def sentences(self, text: str) -> Iterator[str]:
        normalized = self._normalize(text)
        start = 0
        for match in _SENTENCE_BOUNDARY.finditer(normalized):
            chunk = normalized[start : match.start()].strip()
            if chunk:
                yield chunk
            start = match.end()
        remainder = normalized[start:].strip()
        if remainder:
            yield remainder

    def _normalize(self, text: str) -> str:
        # Replace smart quotes/apostrophes with straight ASCII quotes
        normalized = "".join(_APOSTROPHE_VARIANTS.get(ch, ch) for ch in text)
        # Collapse whitespace to keep window sizes predictable
        normalized = re.sub(r"\s+", " ", normalized)
        return normalized.strip()



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Mapping, Protocol, Sequence

from ..models import Conversation, Message


class ConversationFilter(Protocol):
    """Callable object that decides if a conversation should be used."""

    @property
    def name(self) -> str:  # pragma: no cover - trivial
        ...

    def __call__(self, conversation: Conversation, *, year: int) -> bool: ...


def apply_conversation_filters(
    conversations: Mapping[str, Conversation],
    *,
    year: int,
    filters: Sequence[ConversationFilter] | None = None,
) -> dict[str, Conversation]:
    """
    Return a shallow copy of the conversations mapping with only the items that pass
    every provided filter. Filters can be omitted to keep all conversations.
    """

    if not filters:
        return dict(conversations)

    filtered: dict[str, Conversation] = {}
    for key, conversation in conversations.items():
        if all(filter_fn(conversation, year=year) for filter_fn in filters):
            filtered[key] = conversation
    return filtered


def _messages_in_year(conversation: Conversation, year: int) -> Iterable[Message]:
    for message in conversation.messages:
        if getattr(message, "is_context_only", False):
            continue
        if message.timestamp.year != year:
            continue
        yield message


@dataclass
class _BaseFilter:
    min_messages_required: int = 1

    def _count_messages(self, conversation: Conversation, year: int) -> tuple[int, int]:
        sent = 0
        received = 0
        for msg in _messages_in_year(conversation, year):
            if msg.is_from_me:
                sent += 1
            else:
                received += 1
        return sent, received

    def _total_messages(self, conversation: Conversation, year: int) -> int:
        return sum(1 for _ in _messages_in_year(conversation, year))


@dataclass
class _ReceivedToSentRatioFilter(_BaseFilter):
    max_ratio: float = 9.0

    @property
    def name(self) -> str:
        return "received_to_sent_ratio"

    def __call__(self, conversation: Conversation, *, year: int) -> bool:
        sent, received = self._count_messages(conversation, year)

        if sent < self.min_messages_required:
            return False

        if received == 0:
            return True

        ratio = received / max(sent, 1)
        return ratio <= self.max_ratio


@dataclass
class _MinimumResponsesFilter(_BaseFilter):
    min_user_responses: int = 2

    @property
    def name(self) -> str:
        return "minimum_user_responses"

    def __call__(self, conversation: Conversation, *, year: int) -> bool:
        sent, _ = self._count_messages(conversation, year)
        return sent >= self.min_user_responses


@dataclass
class _MinimumTotalMessagesFilter(_BaseFilter):
    min_total_messages: int = 10

    @property
    def name(self) -> str:
        return "minimum_total_messages"

    def __call__(self, conversation: Conversation, *, year: int) -> bool:
        return self._total_messages(conversation, year) >= self.min_total_messages


def received_to_sent_ratio_filter(
    *,
    max_ratio: float = 9.0,
    min_messages_required: int = 1,
) -> ConversationFilter:
    """Factory for the ratio filter to keep instantiation call sites tidy."""

    return _ReceivedToSentRatioFilter(
        max_ratio=max_ratio,
        min_messages_required=min_messages_required,
    )


def minimum_responses_filter(
    *,
    min_user_responses: int = 2,
) -> ConversationFilter:
    """Factory for filtering out conversations with too few outbound replies."""

    return _MinimumResponsesFilter(
        min_user_responses=min_user_responses,
        min_messages_required=min_user_responses,
    )


def minimum_total_messages_filter(
    *,
    min_total_messages: int = 10,
) -> ConversationFilter:
    """Filter out lightweight chats with too little context."""

    return _MinimumTotalMessagesFilter(
        min_total_messages=min_total_messages,
        min_messages_required=min_total_messages,
    )
